from bclearer_core.common_knowledge.convention_shift_operation_types import (
    ConventionShiftOperationTypes,
)
from bclearer_core.configurations.convention_shift_operation_configurations import (
    ConventionShiftOperationConfigurations,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.attribute_names_to_name_objects_convention_shifter import (
    shift_convention_attribute_names_to_name_objects,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.digitalisation_levels.universe_to_semantically_grounded_convention_shifter import (
    shift_convention_universe_to_semantically_grounded,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.generalise_names.generalise_names_convention_shifter import (
    shift_convention_generalise_names,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.objects_to_classes_convention_shifter import (
    shift_convention_objects_to_classes,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_instances_and_exemplars.separate_bespoke_instances_and_exemplars_convention_shifter import (
    shift_convention_separate_bespoke_instances_and_exemplars,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_instances_and_exemplars.separate_standard_instances_and_exemplars_convention_shifter import (
    shift_convention_separate_standard_instances_and_exemplars,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_names_and_instances.separate_bespoke_names_and_instances_convention_shifter import (
    shift_convention_separate_bespoke_names_and_instances,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_names_and_instances.separate_standard_names_and_instances_convention_shifter import (
    shift_convention_separate_standard_names_and_instances,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.uml_names_to_named_objects_convention_shifter import (
    shift_convention_uml_names_to_named_objects,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.session.orchestrators.ea_tools_session_managers import (
    EaToolsSessionManagers,
)
from bclearer_interop_services.ea_interop_service.session.processes.creators.empty_nf_ea_com_universe_creator import (
    create_empty_nf_ea_universe,
)


class ConventionShiftOperationsSubstages:
    def __init__(
        self,
        ea_tools_session_manager: EaToolsSessionManagers,
        convention_shift_operation_configuration: ConventionShiftOperationConfigurations,
        content_universe: NfEaComUniverses,
    ):
        self.ea_tools_session_manager = (
            ea_tools_session_manager
        )

        self.convention_shift_operation_configuration = convention_shift_operation_configuration

        self.content_universe = (
            content_universe
        )

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass

    def run(self) -> NfEaComUniverses:
        output_universe = create_empty_nf_ea_universe(
            ea_tools_session_manager=self.ea_tools_session_manager,
            short_name=self.convention_shift_operation_configuration.output_universe_short_name,
        )

        if (
            self.convention_shift_operation_configuration.operation_type
            == ConventionShiftOperationTypes.OBJECTS_TO_CLASSES
        ):
            shift_convention_objects_to_classes(
                content_universe=self.content_universe,
                output_universe=output_universe,
            )

        elif (
            self.convention_shift_operation_configuration.operation_type
            == ConventionShiftOperationTypes.UML_NAME_TO_NAMED_OBJECT
        ):
            shift_convention_uml_names_to_named_objects(
                content_universe=self.content_universe,
                list_of_configuration_objects=self.convention_shift_operation_configuration.list_of_configuration_objects,
                output_universe=output_universe,
            )

        elif (
            self.convention_shift_operation_configuration.operation_type
            == ConventionShiftOperationTypes.ATTRIBUTE_NAME_TO_NAMED_OBJECT
        ):
            shift_convention_attribute_names_to_name_objects(
                content_universe=self.content_universe,
                list_of_configuration_objects=self.convention_shift_operation_configuration.list_of_configuration_objects,
                output_universe=output_universe,
            )

        elif (
            self.convention_shift_operation_configuration.operation_type
            == ConventionShiftOperationTypes.GENERALISE_NAMES
        ):
            shift_convention_generalise_names(
                content_universe=self.content_universe,
                list_of_configuration_objects=self.convention_shift_operation_configuration.list_of_configuration_objects,
                output_universe=output_universe,
            )

        elif (
            self.convention_shift_operation_configuration.operation_type
            == ConventionShiftOperationTypes.SEPARATE_STANDARD_NAMES_AND_INSTANCES
        ):
            shift_convention_separate_standard_names_and_instances(
                content_universe=self.content_universe,
                output_universe=output_universe,
                package_name=self.convention_shift_operation_configuration.package_name,
            )

        elif (
            self.convention_shift_operation_configuration.operation_type
            == ConventionShiftOperationTypes.SEPARATE_BESPOKE_NAMES_AND_INSTANCES
        ):
            shift_convention_separate_bespoke_names_and_instances(
                content_universe=self.content_universe,
                list_of_configuration_objects=self.convention_shift_operation_configuration.list_of_configuration_objects,
                output_universe=output_universe,
                package_name=self.convention_shift_operation_configuration.package_name,
            )

        elif (
            self.convention_shift_operation_configuration.operation_type
            == ConventionShiftOperationTypes.SEPARATE_STANDARD_INSTANCES_AND_EXEMPLARS
        ):
            shift_convention_separate_standard_instances_and_exemplars(
                content_universe=self.content_universe,
                output_universe=output_universe,
                package_name=self.convention_shift_operation_configuration.package_name,
            )

        elif (
            self.convention_shift_operation_configuration.operation_type
            == ConventionShiftOperationTypes.SEPARATE_BESPOKE_INSTANCES_AND_EXEMPLARS
        ):
            shift_convention_separate_bespoke_instances_and_exemplars(
                content_universe=self.content_universe,
                list_of_configuration_objects=self.convention_shift_operation_configuration.list_of_configuration_objects,
                output_universe=output_universe,
                package_name=self.convention_shift_operation_configuration.package_name,
            )

        elif (
            self.convention_shift_operation_configuration.operation_type
            == ConventionShiftOperationTypes.UNIVERSE_TO_SEMANTICALLY_GROUNDED_DIGITALISATION_LEVEL
        ):
            shift_convention_universe_to_semantically_grounded(
                content_universe=self.content_universe,
                output_universe=output_universe,
            )

        else:
            raise NotImplementedError

        return output_universe



from bclearer_core.objects.verses.nf_verses import (
    NfVerses,
)


class NfMultiverses(NfVerses):
    def __init__(self):
        super().__init__()

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from enum import auto, unique

from bclearer_core.nf.types.column_types import (
    ColumnTypes,
)


@unique
class BclearerAdditionalColumnTypes(
    ColumnTypes,
):
    NAMING_SPACE_NF_UUIDS = auto()

    NAME_INSTANCE_TYPE_NF_UUIDS = auto()

    OWNING_OBJECT_NAMES = auto()

    def __column_name(self) -> str:
        column_name = (
            column_name_mapping[self]
        )

        return column_name

    column_name = property(
        fget=__column_name,
    )


column_name_mapping = {
    BclearerAdditionalColumnTypes.NAMING_SPACE_NF_UUIDS: "naming_space_nf_uuids",
    BclearerAdditionalColumnTypes.NAME_INSTANCE_TYPE_NF_UUIDS: "name_instance_type_nf_uuids",
    BclearerAdditionalColumnTypes.OWNING_OBJECT_NAMES: "owning_object_names",
}



from bclearer_core.constants.standard_constants import (
    DEFAULT_NULL_VALUE,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.common.ea_guid_from_nf_uuid_creator import (
    create_ea_guid_from_nf_uuid,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_association_direction_types import (
    EaAssociationDirectionTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)
from bclearer_interop_services.tuple_service.tuple_attribute_value_getter import (
    get_tuple_attribute_value_if_required,
)
from bclearer_orchestration_services.identification_services.uuid_service.uuid_helpers.uuid_factory import (
    create_new_uuid,
)


def create_association_from_attribute(
    attribute_to_convert_tuple: tuple,
    direction: EaAssociationDirectionTypes,
    type_nf_uuid: str,
    attribute_name: str = DEFAULT_NULL_VALUE,
) -> dict:
    association_nf_uuid = (
        create_new_uuid()
    )

    association_ea_guid = (
        create_ea_guid_from_nf_uuid(
            nf_uuid=association_nf_uuid,
        )
    )

    association_owner_nf_uuid = get_tuple_attribute_value_if_required(
        owning_tuple=attribute_to_convert_tuple,
        attribute_name=NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name,
    )

    if (
        direction
        == EaAssociationDirectionTypes.FORWARD
    ):
        association_client_nf_uuid = (
            type_nf_uuid
        )

        association_supplier_nf_uuid = (
            association_owner_nf_uuid
        )

    else:
        association_client_nf_uuid = (
            association_owner_nf_uuid
        )

        association_supplier_nf_uuid = (
            type_nf_uuid
        )

    association_client_cardinality = __get_connector_client_cardinality(
        attribute_to_convert_tuple=attribute_to_convert_tuple,
    )

    ea_association_dictionary = {
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name: association_ea_guid,
        NfColumnTypes.NF_UUIDS.column_name: association_nf_uuid,
        NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name: association_supplier_nf_uuid,
        NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS.column_name: association_client_nf_uuid,
        NfEaComColumnTypes.CONNECTORS_DIRECTION_TYPE_NAME.column_name: "Source -> Destination",
        NfEaComColumnTypes.CONNECTORS_ELEMENT_TYPE_NAME.column_name: EaConnectorTypes.ASSOCIATION.type_name,
        NfEaComColumnTypes.CONNECTORS_SOURCE_CARDINALITY.column_name: DEFAULT_NULL_VALUE,
        NfEaComColumnTypes.CONNECTORS_DEST_CARDINALITY.column_name: association_client_cardinality,
        NfEaComColumnTypes.STEREOTYPEABLE_OBJECTS_EA_OBJECT_STEREOTYPES.column_name: [],
        NfEaComColumnTypes.REPOSITORIED_OBJECTS_EA_REPOSITORY.column_name: DEFAULT_NULL_VALUE,
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name: attribute_name,
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NOTES.column_name: DEFAULT_NULL_VALUE,
    }

    return ea_association_dictionary


def __get_connector_client_cardinality(
    attribute_to_convert_tuple: tuple,
) -> str:
    attribute_lower_bound = get_tuple_attribute_value_if_required(
        owning_tuple=attribute_to_convert_tuple,
        attribute_name=NfEaComColumnTypes.ATTRIBUTES_LOWER_BOUNDS.column_name,
    )

    attribute_upper_bound = get_tuple_attribute_value_if_required(
        owning_tuple=attribute_to_convert_tuple,
        attribute_name=NfEaComColumnTypes.ATTRIBUTES_UPPER_BOUNDS.column_name,
    )

    if (
        attribute_lower_bound
        != attribute_upper_bound
    ):
        connector_client_cardinality = (
            attribute_lower_bound
            + ".."
            + attribute_upper_bound
        )

    else:
        connector_client_cardinality = (
            attribute_lower_bound
        )

    return connector_client_cardinality



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)


def get_nf_uuid_from_ea_guid_from_collection(
    nf_ea_com_universe: NfEaComUniverses,
    collection_type: NfEaComCollectionTypes,
    ea_guid: str,
) -> str:
    collection = nf_ea_com_universe.nf_ea_com_registry.dictionary_of_collections[
        collection_type
    ]

    nf_uuid = collection.at[
        collection[
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
        ]
        .eq(ea_guid)
        .idxmax(),
        NfColumnTypes.NF_UUIDS.column_name,
    ]

    return nf_uuid



from bclearer_core.common_knowledge.digitialisation_level_stereotype_matched_ea_objects import (
    DigitalisationLevelStereotypeMatchedEaObjects,
)


class LoadHdf5ModelConfigurations:
    def __init__(
        self,
        resource_namespace: str,
        resource_file_name: str,
        universe_short_name: str,
        default_digitalisation_level_stereotype: DigitalisationLevelStereotypeMatchedEaObjects = None,
        ea_guid_to_digitalisation_level_stereotype_dictionary: dict = None,
    ):
        self.resource_namespace = (
            resource_namespace
        )

        self.resource_file_name = (
            resource_file_name
        )

        self.universe_short_name = (
            universe_short_name
        )

        self.default_digitalisation_level_stereotype = default_digitalisation_level_stereotype

        self.ea_guid_to_digitalisation_level_stereotype_dictionary = ea_guid_to_digitalisation_level_stereotype_dictionary

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



"""Unix specific helpers for user session identifiers."""

import os


def get_unix_session_id() -> int:
    if hasattr(os, "getsid"):
        try:
            return os.getsid(0)
        except OSError:
            pass

    return os.getpid()



from bclearer_core.substages.operations.common.classifier_adder import (
    add_new_classifier_to_dictionary,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_element_types import (
    EaElementTypes,
)


def add_new_association_class_to_dictionary(
    new_classifier_dictionary: dict,
    package_nf_uuid: str,
) -> str:
    association_class_nf_uuid = add_new_classifier_to_dictionary(
        new_classifier_dictionary=new_classifier_dictionary,
        package_nf_uuid=package_nf_uuid,
        ea_element_type=EaElementTypes.ASSOCIATION,
    )

    return association_class_nf_uuid



"""Unix specific account context helpers."""

import os
from typing import Tuple

from bclearer_core.infrastructure.session.common.session_config import (
    UNIX_ADMIN_ACCOUNT_TYPE,
    UNIX_DEFAULT_DOMAIN_NAME,
    UNIX_STANDARD_ACCOUNT_TYPE,
)

try:
    import pwd
except (
    ImportError
):  # pragma: no cover - platforms without pwd
    pwd = None


def get_unix_account_context() -> (
    Tuple[str, str]
):
    if pwd:
        try:
            user_info = pwd.getpwuid(
                os.getuid()
            )
        except KeyError:
            user_info = None
    else:
        user_info = None

    full_name = ""
    if user_info and user_info.pw_gecos:
        full_name = (
            user_info.pw_gecos.split(
                ","
            )[0].strip()
        )

    domain_name = (
        full_name
        or UNIX_DEFAULT_DOMAIN_NAME
    )
    account_type = (
        UNIX_ADMIN_ACCOUNT_TYPE
        if os.getuid() == 0
        else UNIX_STANDARD_ACCOUNT_TYPE
    )

    return domain_name, account_type



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.b_evolve.adjustment_operations.convert_attributes_to_associations.association_from_attribute_creator import (
    create_association_from_attribute,
)
from bclearer_core.substages.operations.b_evolve.adjustment_operations.convert_attributes_to_associations.dependency_from_attribute_creator import (
    create_dependency_from_attribute,
)
from bclearer_core.substages.operations.b_evolve.adjustment_operations.convert_attributes_to_associations.type_from_attribute_creator import (
    create_type_from_attribute,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_association_direction_types import (
    EaAssociationDirectionTypes,
)


# TODO: added to patch boson pipeline, not there in latest master from BORO.
def convert_attribute(
    attribute_to_convert_tuple: tuple,
    new_classifiers_dictionary: dict,
    new_connectors_dictionary: dict,
    direction: EaAssociationDirectionTypes,
    package_nf_uuid: str,
) -> tuple:
    (
        type_nf_uuid,
        new_classifiers_dictionary,
    ) = __add_type(
        attribute_to_convert_tuple=attribute_to_convert_tuple,
        package_nf_uuid=package_nf_uuid,
        new_classifiers_dictionary=new_classifiers_dictionary,
    )

    new_connectors_dictionary = __add_association(
        attribute_to_convert_tuple=attribute_to_convert_tuple,
        direction=direction,
        type_nf_uuid=type_nf_uuid,
        new_connectors_dictionary=new_connectors_dictionary,
    )

    new_connectors_dictionary = __add_dependency(
        attribute_to_convert_tuple=attribute_to_convert_tuple,
        type_nf_uuid=type_nf_uuid,
        new_connectors_dictionary=new_connectors_dictionary,
    )

    return (
        new_classifiers_dictionary,
        new_connectors_dictionary,
    )


def __add_type(
    attribute_to_convert_tuple: tuple,
    package_nf_uuid: str,
    new_classifiers_dictionary: dict,
) -> tuple:
    type_dictionary = create_type_from_attribute(
        attribute_to_convert_tuple=attribute_to_convert_tuple,
        package_nf_uuid=package_nf_uuid,
    )

    type_nf_uuid = type_dictionary[
        NfColumnTypes.NF_UUIDS.column_name
    ]

    new_classifiers_dictionary[
        type_nf_uuid
    ] = type_dictionary

    return (
        type_nf_uuid,
        new_classifiers_dictionary,
    )


def __add_association(
    attribute_to_convert_tuple: tuple,
    direction: EaAssociationDirectionTypes,
    type_nf_uuid: str,
    new_connectors_dictionary: dict,
) -> dict:
    association_dictionary = create_association_from_attribute(
        attribute_to_convert_tuple=attribute_to_convert_tuple,
        direction=direction,
        type_nf_uuid=type_nf_uuid,
    )

    association_nf_uuid = association_dictionary[
        NfColumnTypes.NF_UUIDS.column_name
    ]

    new_connectors_dictionary[
        association_nf_uuid
    ] = association_dictionary

    return new_connectors_dictionary


def __add_dependency(
    attribute_to_convert_tuple: tuple,
    type_nf_uuid: str,
    new_connectors_dictionary: dict,
) -> dict:
    generalisation_dictionary = create_dependency_from_attribute(
        attribute_to_convert_tuple=attribute_to_convert_tuple,
        type_nf_uuid=type_nf_uuid,
    )

    association_nf_uuid = generalisation_dictionary[
        NfColumnTypes.NF_UUIDS.column_name
    ]

    new_connectors_dictionary[
        association_nf_uuid
    ] = generalisation_dictionary

    return new_connectors_dictionary



import pandas

from bclearer_core.objects.registers.registers import Registers


class PandasTableRegisters(Registers[pandas.DataFrame]):
    def __init__(
            self,
            owning_registry,
            name: str,
            content: pandas.DataFrame | None = None):
        super().__init__(
            owning_registry=owning_registry,
            name=name,
            content=content)
        
    def get_model(
            self):
        pandas_dataframe = \
            self.get_content()
        
        return \
            {
                "table_name": self.name,
                "num_columns": len(pandas_dataframe.columns),
                "num_rows": len(pandas_dataframe),
                "columns": [
                    {
                        "name": col,
                        "dtype": str(pandas_dataframe[col].dtype),
                        "nullable": bool(pandas_dataframe[col].isnull().any())
                    }
                    for col in pandas_dataframe.columns
                ]
            }
        
    def get_content(
            self) \
            -> pandas.DataFrame:
        content = \
            super().get_content()
        
        return \
            content.copy()
    
    def export_to_disk(
            self, 
            base_folder_path: str):
        pandas_dataframe = \
            self.get_content()
        
        # TODO: This should support multiple formats!
        file_path = \
            f"{base_folder_path}/{self.name}.csv"
        
        pandas_dataframe.to_csv(
            file_path,
            index=False)
        
    def __str__(self) -> str:
        return \
            f"PandasTableRegisterLatest(name={self.name}, rows={len(self._content) if self._content is not None else 'None'})"



import uuid

from bclearer_core.infrastructure.session.common.enums.operating_system_kind import (
    OperatingSystemKind,
)
from bclearer_core.infrastructure.session.common.platform_type_getter import (
    get_operating_system_kind,
)
from bclearer_core.infrastructure.session.common.session_config import (
    UNKNOWN_MACHINE_FALLBACK_PREFIX,
)
from bclearer_core.infrastructure.session.operating_system.helpers.linux.linux_machine_identifier import (
    get_linux_machine_identifier,
)
from bclearer_core.infrastructure.session.operating_system.helpers.macos.macos_machine_identifier import (
    get_macos_machine_identifier,
)


def get_machine_sid() -> str:
    operating_system_kind = (
        get_operating_system_kind()
    )

    if (
        operating_system_kind
        is OperatingSystemKind.WINDOWS
    ):
        from bclearer_core.infrastructure.session.operating_system.helpers.windows.windows_machine_sid import (
            get_windows_machine_sid,
        )

        return get_windows_machine_sid()

    if (
        operating_system_kind
        is OperatingSystemKind.LINUX
    ):
        return (
            get_linux_machine_identifier()
        )

    if (
        operating_system_kind
        is OperatingSystemKind.MACOS
    ):
        return (
            get_macos_machine_identifier()
        )

    return f"{UNKNOWN_MACHINE_FALLBACK_PREFIX}{uuid.getnode():012x}"



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.common.ea_guid_from_nf_uuid_creator import (
    create_ea_guid_from_nf_uuid,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_element_types import (
    EaElementTypes,
)
from bclearer_interop_services.tuple_service.tuple_attribute_value_getter import (
    get_tuple_attribute_value_if_required,
)
from bclearer_orchestration_services.identification_services.uuid_service.uuid_helpers.uuid_factory import (
    create_new_uuid,
)


def create_type_from_attribute(
    attribute_to_convert_tuple: tuple,
    package_nf_uuid: str,
) -> dict:
    new_type_nf_uuid = create_new_uuid()

    new_type_ea_guid = (
        create_ea_guid_from_nf_uuid(
            nf_uuid=new_type_nf_uuid,
        )
    )

    new_type_name = get_tuple_attribute_value_if_required(
        owning_tuple=attribute_to_convert_tuple,
        attribute_name=NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name,
    )

    ea_classifier_dictionary = {
        NfColumnTypes.NF_UUIDS.column_name: new_type_nf_uuid,
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name: new_type_ea_guid,
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name: new_type_name,
        NfEaComColumnTypes.ELEMENTS_EA_OBJECT_TYPE.column_name: EaElementTypes.CLASS.type_name,
        NfEaComColumnTypes.PACKAGEABLE_OBJECTS_PARENT_EA_ELEMENT.column_name: package_nf_uuid,
    }

    return ea_classifier_dictionary



"""Template for conftest.py."""

CONFTEST_TEMPLATE = """import pytest


@pytest.fixture(scope="module")
def e2e_test_setup():
    # Add setup code here
    pass


@pytest.fixture(scope="module")
def e2e_test_teardown():
    # Add teardown code here
    pass
"""



from enum import auto

from bclearer_orchestration_services.identification_services.b_identity_ecosystem.common_knowledge.bie_infrastructure_types import (
    BieInfrastructureTypes,
)


class BieOsUserProfileSessionTypes(
    BieInfrastructureTypes
):
    NOT_SET = auto()

    BIE_OS_USER_PROFILE_SESSION = auto()



"""Domain universe container providing access to specialised registries."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from bclearer_interop_services.graph_services.b_simple_graph_service.objects.b_simple_graphs_universe_registries import (
    BSimpleGraphsUniverseRegistries,
)
from bclearer_interop_services.file_system_service.objects.folders import (
    Folders,
)


class BClearerDomainUniverses:
    """Aggregate for the various registry types used within a universe."""

    def __init__(self) -> None:
        self._graph_registries = BSimpleGraphsUniverseRegistries(
            owning_b_simple_graphs_universe=self,
        )

    # ------------------------------------------------------------------
    # Graph data object helpers
    # ------------------------------------------------------------------
    def add_graph_data_object(
        self,
        registry_name: str,
        register_name: str,
        graph_name: str,
        graph_object: Any,
    ) -> None:
        """Persist a graph object for the specified registry/register pair."""

        self._graph_registries.add_graph(
            registry_name=registry_name,
            register_name=register_name,
            graph_name=graph_name,
            graph_object=graph_object,
        )

    def remove_graph_data_object(
        self,
        registry_name: str,
        register_name: str,
        graph_name: str,
    ) -> None:
        """Remove a previously stored graph object."""

        self._graph_registries.remove_graph(
            registry_name=registry_name,
            register_name=register_name,
            graph_name=graph_name,
        )

    def list_graph_data_objects(
        self,
        registry_name: str,
        register_name: str,
    ) -> Mapping[str, Any]:
        """Return the graph objects stored for a register."""

        return self._graph_registries.list_graphs(
            registry_name=registry_name,
            register_name=register_name,
        )

    def export_graph_data_objects_to_graph_ml(
        self,
        output_folder: Folders,
    ) -> None:
        """Export all stored graphs to GraphML format."""

        self._graph_registries.export_all_b_simple_graphs_to_graph_ml(
            output_folder=output_folder,
        )




from bclearer_core.configurations.content_operation_configurations import (
    ContentOperationConfigurations,
)
from bclearer_core.configurations.load_hdf5_model_configurations import (
    LoadHdf5ModelConfigurations,
)
from bclearer_core.substages.operations.a_load.content_operations.runners.hdf5_to_content_universe_loader import (
    load_hdf5_model_to_content_universe,
)
from bclearer_core.substages.operations.b_evolve.content_operations.content_operations_substages import (
    ContentOperationsSubstages,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.session.orchestrators.ea_tools_session_managers import (
    EaToolsSessionManagers,
)


def run_merge_hdf5_model_content_operations_substage(
    ea_tools_session_manager: EaToolsSessionManagers,
    content_universe_from_input: NfEaComUniverses,
    load_hdf5_model_configuration: LoadHdf5ModelConfigurations,
    content_operation_configuration: ContentOperationConfigurations,
) -> NfEaComUniverses:
    content_universe_from_hdf5 = load_hdf5_model_to_content_universe(
        ea_tools_session_manager=ea_tools_session_manager,
        load_hdf5_model_configuration=load_hdf5_model_configuration,
    )

    with ContentOperationsSubstages(
        ea_tools_session_manager=ea_tools_session_manager,
        content_1_universe=content_universe_from_input,
        content_2_universe=content_universe_from_hdf5,
        content_operation_configuration=content_operation_configuration,
    ) as content_operations_substage:
        content_operations_substage_output_universe = (
            content_operations_substage.run()
        )

        return content_operations_substage_output_universe



from datetime import datetime

from bclearer_core.constants.standard_constants import (
    HR_DATE_TIME_FORMAT,
    HR_NAME_COMPONENT_DIVIDER,
)
from bclearer_core.infrastructure.session.app_run_session.bie_app_run_session_types import (
    BieAppRunSessionTypes,
)
from bclearer_core.infrastructure.session.common.session_objects import (
    SessionObjects,
)
from bclearer_core.infrastructure.session.os_user_session.os_user_profile_session_objects import (
    OsUserProfileSessionObjects,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.bie_id_creation_module.bie_id_creation_facade import (
    BieIdCreationFacade,
)


class AppRunSessionObjects(
    SessionObjects
):
    def __init__(
        self,
        bie_app_run_session_type: BieAppRunSessionTypes,
    ):  # TODO: needs to be integrated (called by) into the b_application_runner
        self.bie_app_run_session_type = (
            bie_app_run_session_type
        )

        self.owning_os_user_profile_session = (
            OsUserProfileSessionObjects()
        )

        # TODO: the app run session timestamp is not the same date time than the os user session
        self.app_run_session_date_time = (
            datetime.now()
        )

        # TODO: Design a bie id - get to work with other OSs
        bie_id = BieIdCreationFacade.create_order_sensitive_bie_id_for_multiple_objects(
            input_objects=[
                bie_app_run_session_type.item_bie_identity,
                self.owning_os_user_profile_session.bie_id,
                self.app_run_session_date_time,
            ]
        )

        base_hr_name = (
            self.owning_os_user_profile_session.owning_operating_system_object.base_hr_name
            + HR_NAME_COMPONENT_DIVIDER
            + self.owning_os_user_profile_session.owning_os_user_profile.base_hr_name
            + HR_NAME_COMPONENT_DIVIDER
            + self.app_run_session_date_time.strftime(
                format=HR_DATE_TIME_FORMAT
            )
        )

        super().__init__(
            bie_id=bie_id,
            base_hr_name=base_hr_name,
            bie_infrastructure_type=bie_app_run_session_type,
        )



from bclearer_core.infrastructure.session.app_run_session.app_run_session_objects import (
    AppRunSessionObjects,
)
from bclearer_core.infrastructure.session.bie_id_registerers.app_run_session_objects_registerer import (
    register_app_run_session_objects,
)
from bclearer_core.infrastructure.session.bie_id_registerers.operating_system_objects_registerer import (
    register_operating_system_objects,
)
from bclearer_core.infrastructure.session.bie_id_registerers.os_user_profile_objects_registerer import (
    register_os_user_profile_objects,
)
from bclearer_core.infrastructure.session.bie_id_registerers.os_user_profile_session_objects_registerer import (
    register_os_user_profile_session_objects,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.infrastructure.registrations.bie_infrastructure_registries import (
    BieInfrastructureRegistries,
)


def register_infrastructure_bie_ids(
    bie_infrastructure_registry: BieInfrastructureRegistries,
    app_run_session_object: AppRunSessionObjects,
) -> None:
    register_app_run_session_objects(
        bie_infrastructure_registry=bie_infrastructure_registry,
        app_run_session_object=app_run_session_object,
    )

    register_operating_system_objects(
        bie_infrastructure_registry=bie_infrastructure_registry
    )

    register_os_user_profile_objects(
        bie_infrastructure_registry=bie_infrastructure_registry
    )

    register_os_user_profile_session_objects(
        bie_infrastructure_registry=bie_infrastructure_registry
    )



from bclearer_orchestration_services.identification_services.b_identity_ecosystem.common_knowledge.bie_types import (
    BieTypes,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.objects.bie_ids import (
    BieIds,
)


class BieObjects:
    def __init__(
        self,
        bie_id: BieIds,
        base_hr_name: str,
        bie_type: BieTypes,
    ):
        self.bie_id = bie_id

        self.base_hr_name = base_hr_name

        self.bie_type = bie_type

    def __hash__(self):
        return hash(
            self.bie_id.int_value
        )



from bclearer_core.configurations.b_clearer_configurations.b_clearer_domain_configurations import (
    BClearerDomainConfigurations,
)


class BClearerDomainGraphConfigurations(
    BClearerDomainConfigurations
):
    SOURCE_GRAPH_FILE_FORMAT = str()



import csv
import os

from bclearer_core.common_knowledge.digitialisation_level_stereotype_matched_ea_objects import (
    DigitalisationLevelStereotypeMatchedEaObjects,
)
from bclearer_core.configurations.run_configurations import (
    RunConfigurations,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.common.nf_uuid_from_ea_guid_from_collection_getter import (
    get_nf_uuid_from_ea_guid_from_collection,
)
from bclearer_interop_services.dataframe_service.dataframe_helpers.dataframe_filter_and_renamer import (
    dataframe_filter_and_rename,
)
from bclearer_interop_services.dataframe_service.dataframe_helpers.dataframe_mergers import (
    left_merge_dataframes,
)
from bclearer_interop_services.delimited_text.dataframe_dictionary_to_csv_files_writer import (
    write_dataframe_dictionary_to_csv_files,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.domain_migration.nf_ea_com_to_domain_migration.processes.nf_ea_com_to_standard_tables_dictionary_converter import (
    convert_nf_ea_com_to_standard_tables_dictionary,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.column_types.nf_domains.standard_object_table_column_types import (
    StandardObjectTableColumnTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_element_types import (
    EaElementTypes,
)
from bclearer_interop_services.file_system_service.objects.files import (
    Files,
)
from bclearer_interop_services.tuple_service.tuple_attribute_value_getter import (
    get_tuple_attribute_value_if_required,
)
from bclearer_orchestration_services.log_environment_utility_service.common_knowledge.constants import (
    NAME_VALUE_DELIMITER,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)
from bclearer_orchestration_services.reporting_service.wrappers.run_and_log_function_wrapper import (
    run_and_log_function,
)


@run_and_log_function
def instrument_and_visualize(
    output_folder_name: str,
    visualization_substage_output_universe: NfEaComUniverses,
):
    __report_summary(
        visualization_substage_output_universe=visualization_substage_output_universe,
        output_folder_name=output_folder_name,
    )

    __report_digitalisation_levels_population(
        visualization_substage_output_universe=visualization_substage_output_universe,
        output_folder_name=output_folder_name,
    )

    if RunConfigurations.hdf5_output:
        __write_hdf5_file(
            output_folder_name=output_folder_name,
            visualization_substage_output_universe=visualization_substage_output_universe,
        )

    if RunConfigurations.csv_output:
        __export_to_csv(
            output_folder_name=output_folder_name,
            visualization_substage_output_universe=visualization_substage_output_universe,
        )


def __report_summary(
    visualization_substage_output_universe: NfEaComUniverses,
    output_folder_name: str,
) -> None:
    visualization_substage_output_universe.nf_ea_com_registry.create_or_update_nf_ea_com_summary_table()

    visualization_substage_identifier_code = (
        visualization_substage_output_universe.ea_repository.short_name
    )

    summary_table_by_type = visualization_substage_output_universe.nf_ea_com_registry.dictionary_of_collections[
        NfEaComCollectionTypes.SUMMARY_TABLE_BY_TYPE
    ]

    summary_table_by_type_filtered = summary_table_by_type.loc[
        (
            summary_table_by_type[
                "minor_types"
            ]
            == EaConnectorTypes.ASSOCIATION.type_name
        )
        | (
            summary_table_by_type[
                "minor_types"
            ]
            == "Attribute"
        )
        | (
            summary_table_by_type[
                "minor_types"
            ]
            == EaConnectorTypes.DEPENDENCY.type_name
        )
        | (
            summary_table_by_type[
                "minor_types"
            ]
            == EaConnectorTypes.GENERALIZATION.type_name
        )
        | (
            summary_table_by_type[
                "minor_types"
            ]
            == EaElementTypes.CLASS.type_name
        )
        | (
            summary_table_by_type[
                "minor_types"
            ]
            == EaElementTypes.OBJECT.type_name
        )
        | (
            summary_table_by_type[
                "minor_types"
            ]
            == "Stereotype_usage"
        )
    ]

    visualization_substage_folder_path = os.path.join(
        output_folder_name,
        visualization_substage_output_universe.ea_repository.short_name,
    )

    os.mkdir(
        visualization_substage_folder_path,
    )

    summary_file_full_path = os.path.join(
        visualization_substage_folder_path,
        visualization_substage_output_universe.ea_repository.short_name
        + "_summary.csv",
    )

    summary_table_by_type_filtered.to_csv(
        path_or_buf=summary_file_full_path,
        sep=",",
        quotechar='"',
        index=False,
        quoting=csv.QUOTE_ALL,
    )

    for (
        summary_tuple
    ) in (
        summary_table_by_type_filtered.itertuples()
    ):
        main_type = get_tuple_attribute_value_if_required(
            owning_tuple=summary_tuple,
            attribute_name="main_types",
        )
        minor_type = get_tuple_attribute_value_if_required(
            owning_tuple=summary_tuple,
            attribute_name="minor_types",
        )
        row_count = get_tuple_attribute_value_if_required(
            owning_tuple=summary_tuple,
            attribute_name="row_count",
        )

        message_list = [
            "@count@",
            visualization_substage_identifier_code,
            main_type,
            minor_type,
            str(row_count),
        ]

        log_message(
            message=NAME_VALUE_DELIMITER.join(
                message_list,
            ),
        )


def __report_digitalisation_levels_population(
    visualization_substage_output_universe: NfEaComUniverses,
    output_folder_name: str,
) -> None:
    output_universe_stereotype_usages = visualization_substage_output_universe.nf_ea_com_registry.dictionary_of_collections[
        NfEaComCollectionTypes.STEREOTYPE_USAGE
    ]

    digitalisation_level_guids_set = (
        DigitalisationLevelStereotypeMatchedEaObjects.get_ea_guids()
    )

    digitalisation_level_nf_uuids = (
        list()
    )

    for (
        digitalisation_level_guid
    ) in digitalisation_level_guids_set:
        digitalisation_level_nf_uuids.append(
            get_nf_uuid_from_ea_guid_from_collection(
                nf_ea_com_universe=visualization_substage_output_universe,
                collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
                ea_guid=digitalisation_level_guid,
            ),
        )

    digitalisation_levels_stereotype_usages = output_universe_stereotype_usages[
        output_universe_stereotype_usages[
            "stereotype_nf_uuids"
        ].isin(
            digitalisation_level_nf_uuids,
        )
    ]

    digitalisation_levels_stereotype_usages_summary = digitalisation_levels_stereotype_usages.groupby(
        ["stereotype_nf_uuids"],
    ).count()

    digitalisation_levels_stereotype_usages_summary[
        "stereotype_nf_uuids"
    ] = (
        digitalisation_levels_stereotype_usages_summary.index
    )

    digitalisation_levels_population_table = left_merge_dataframes(
        master_dataframe=digitalisation_levels_stereotype_usages_summary,
        master_dataframe_key_columns=[
            "stereotype_nf_uuids",
        ],
        merge_suffixes=["1", "2"],
        foreign_key_dataframe=visualization_substage_output_universe.get_ea_stereotypes(),
        foreign_key_dataframe_fk_columns=[
            NfColumnTypes.NF_UUIDS.column_name,
        ],
        foreign_key_dataframe_other_column_rename_dictionary={
            "stereotype_nf_uuids": "stereotype_nf_uuids",
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name: "digitalisation_level_names",
        },
    )

    digitalisation_levels_population_table = dataframe_filter_and_rename(
        dataframe=digitalisation_levels_population_table,
        filter_and_rename_dictionary={
            "digitalisation_level_names": "digitalisation_level_names",
            StandardObjectTableColumnTypes.CLIENT_NF_UUIDS.column_name: "number_of_objects",
        },
    )

    visualization_substage_folder_path = os.path.join(
        output_folder_name,
        visualization_substage_output_universe.ea_repository.short_name,
    )

    digitalisation_levels_population_file_full_path = os.path.join(
        visualization_substage_folder_path,
        visualization_substage_output_universe.ea_repository.short_name
        + "_digitalisation_levels_population.csv",
    )

    digitalisation_levels_population_table.to_csv(
        path_or_buf=digitalisation_levels_population_file_full_path,
        sep=",",
        quotechar='"',
        index=False,
        quoting=csv.QUOTE_ALL,
    )


def __report_dependency_depth(
    visualization_substage_output_universe: NfEaComUniverses,
) -> None:
    visualization_substage_output_universe.nf_ea_com_registry.create_or_update_dependency_depths_table()

    dependency_depths_table = visualization_substage_output_universe.nf_ea_com_registry.dictionary_of_collections[
        NfEaComCollectionTypes.DEPENDENCY_DEPTHS_TABLE
    ]

    log_message(
        message="The type-instance depth of this model is: "
        + str(
            dependency_depths_table[
                "level_depth"
            ].max(),
        ),
    )

    print(dependency_depths_table)


def __write_hdf5_file(
    output_folder_name: str,
    visualization_substage_output_universe: NfEaComUniverses,
) -> None:
    visualization_substage_identifier_code = (
        visualization_substage_output_universe.ea_repository.short_name
    )

    hdf5_directory_path = os.path.join(
        output_folder_name,
        visualization_substage_identifier_code,
    )

    hdf5_file_name = (
        visualization_substage_identifier_code
        + "_hdf5_export.hdf5"
    )

    hdf5_file_path = os.path.join(
        hdf5_directory_path,
        hdf5_file_name,
    )

    hdf5_file = Files(
        absolute_path_string=str(
            hdf5_file_path,
        ),
    )

    visualization_substage_output_universe.nf_ea_com_registry.export_registry_to_hdf5(
        hdf5_file=hdf5_file,
    )


def __export_to_csv(
    output_folder_name: str,
    visualization_substage_output_universe: NfEaComUniverses,
) -> None:
    universe_as_standard_tables_dictionary = convert_nf_ea_com_to_standard_tables_dictionary(
        nf_ea_com_universe=visualization_substage_output_universe,
    )

    visualization_substage_identifier_code = (
        visualization_substage_output_universe.ea_repository.short_name
    )

    csv_directory_path = os.path.join(
        output_folder_name,
        visualization_substage_identifier_code,
        "csv",
    )

    os.mkdir(csv_directory_path)

    write_dataframe_dictionary_to_csv_files(
        csv_files_folder=csv_directory_path,
        dataframe_dictionary=universe_as_standard_tables_dictionary,
    )



from bclearer_core.common_knowledge.adjustment_operation_types import (
    AdjustmentOperationTypes,
)
from bclearer_core.configurations.load_hdf5_model_configurations import (
    LoadHdf5ModelConfigurations,
)
from bclearer_core.configurations.operation_configurations import (
    OperationConfigurations,
)


class AdjustmentOperationConfigurations(
    OperationConfigurations,
):
    def __init__(
        self,
        adjustment_operation_type: AdjustmentOperationTypes,
        adjustment_universe_load_hdf5_model_configuration: LoadHdf5ModelConfigurations,
        output_universe_short_name: str,
    ):
        super().__init__(
            operation_type=adjustment_operation_type,
        )

        self.adjustment_universe_load_hdf5_model_configuration = adjustment_universe_load_hdf5_model_configuration

        self.output_universe_short_name = (
            output_universe_short_name
        )

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from bclearer_core.common_knowledge.bclearer_matched_ea_objects import (
    BclearerMatchedEaObjects,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_names_and_instances.ea_attribute_remover import (
    remove_ea_attributes,
)
from bclearer_core.substages.operations.common.class_adder import (
    add_new_class_to_dictionary,
)
from bclearer_core.substages.operations.common.connector_adder import (
    add_new_connector_to_dictionary,
)
from bclearer_core.substages.operations.common.nf_uuid_from_ea_guid_from_collection_getter import (
    get_nf_uuid_from_ea_guid_from_collection,
)
from bclearer_core.substages.operations.common.stereotype_adder import (
    add_new_stereotype_usage_to_dictionary,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)
from bclearer_interop_services.tuple_service.tuple_attribute_value_getter import (
    get_tuple_attribute_value_if_required,
)
from pandas import DataFrame


def separate_instances_and_exemplars(
    nf_ea_com_universe: NfEaComUniverses,
    name_instance_type_nf_uuid: str,
    ea_attributes: DataFrame,
    new_ea_objects_dictionary: dict,
    package_nf_uuid: str,
    digitalisation_level_stereotype_nf_uuid: str,
):
    ea_classifiers = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_classifiers()
    )

    name_instance_type_name = ea_classifiers.at[
        ea_classifiers[
            NfColumnTypes.NF_UUIDS.column_name
        ]
        .eq(name_instance_type_nf_uuid)
        .idxmax(),
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name,
    ]

    if (
        "Instances"
        in name_instance_type_name
    ):
        name_exemplar_type_name = name_instance_type_name.replace(
            "Instances",
            "Exemplars",
        )

    else:
        name_exemplar_type_name = (
            name_instance_type_name
            + " Exemplars"
        )

    name_exemplar_type_nf_uuid = __create_name_exemplar_type(
        nf_ea_com_universe=nf_ea_com_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
        package_nf_uuid=package_nf_uuid,
        name_exemplar_type_name=name_exemplar_type_name,
        name_instance_type_nf_uuid=name_instance_type_nf_uuid,
    )

    add_new_stereotype_usage_to_dictionary(
        new_stereotype_usage_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.STEREOTYPE_USAGE
        ],
        client_nf_uuid=name_exemplar_type_nf_uuid,
        client_collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
        stereotype_nf_uuid=digitalisation_level_stereotype_nf_uuid,
    )

    exemplified_by_stereotype_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
        ea_guid=BclearerMatchedEaObjects.EXEMPLIFIED_BY_STEREOTYPE.ea_guid,
    )

    name_exemplar_stereotype_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
        ea_guid=BclearerMatchedEaObjects.NAME_EXEMPLAR_STEREOTYPE.ea_guid,
    )

    for (
        ea_attribute_tuple
    ) in ea_attributes.itertuples():
        __create_name_exemplar(
            new_ea_objects_dictionary=new_ea_objects_dictionary,
            package_nf_uuid=package_nf_uuid,
            ea_attribute_tuple=ea_attribute_tuple,
            name_exemplar_type_nf_uuid=name_exemplar_type_nf_uuid,
            exemplified_by_stereotype_nf_uuid=exemplified_by_stereotype_nf_uuid,
            name_exemplar_stereotype_nf_uuid=name_exemplar_stereotype_nf_uuid,
            digitalisation_level_stereotype_nf_uuid=digitalisation_level_stereotype_nf_uuid,
        )

    remove_ea_attributes(
        nf_ea_com_universe=nf_ea_com_universe,
        ea_attributes=ea_attributes,
    )


def __create_name_exemplar_type(
    nf_ea_com_universe: NfEaComUniverses,
    new_ea_objects_dictionary: dict,
    package_nf_uuid: str,
    name_exemplar_type_name: str,
    name_instance_type_nf_uuid: str,
) -> str:
    name_exemplar_type_nf_uuid = add_new_class_to_dictionary(
        new_classifier_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CLASSIFIERS
        ],
        package_nf_uuid=package_nf_uuid,
        class_name=name_exemplar_type_name,
    )

    name_exemplars_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
        ea_guid=BclearerMatchedEaObjects.NAME_EXEMPLARS.ea_guid,
    )

    add_new_connector_to_dictionary(
        new_connector_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CONNECTORS
        ],
        place_1_nf_uuid=name_exemplar_type_nf_uuid,
        place_2_nf_uuid=name_exemplars_nf_uuid,
        connector_type=EaConnectorTypes.GENERALIZATION,
    )

    add_new_connector_to_dictionary(
        new_connector_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CONNECTORS
        ],
        place_1_nf_uuid=name_exemplar_type_nf_uuid,
        place_2_nf_uuid=name_instance_type_nf_uuid,
        connector_type=EaConnectorTypes.ASSOCIATION,
        connector_name=BclearerMatchedEaObjects.EXEMPLIFIED_BY_STEREOTYPE.object_name,
    )

    return name_exemplar_type_nf_uuid


def __create_name_exemplar(
    new_ea_objects_dictionary: dict,
    package_nf_uuid: str,
    ea_attribute_tuple: tuple,
    name_exemplar_type_nf_uuid: str,
    exemplified_by_stereotype_nf_uuid: str,
    name_exemplar_stereotype_nf_uuid: str,
    digitalisation_level_stereotype_nf_uuid: str,
):
    attributed_object_nf_uuid = get_tuple_attribute_value_if_required(
        owning_tuple=ea_attribute_tuple,
        attribute_name=NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name,
    )

    name_exemplar_name = get_tuple_attribute_value_if_required(
        owning_tuple=ea_attribute_tuple,
        attribute_name=NfEaComColumnTypes.ELEMENT_COMPONENTS_DEFAULT.column_name,
    )

    name_exemplar_nf_uuid = add_new_class_to_dictionary(
        new_classifier_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CLASSIFIERS
        ],
        package_nf_uuid=package_nf_uuid,
        class_name=name_exemplar_name,
    )

    add_new_stereotype_usage_to_dictionary(
        new_stereotype_usage_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.STEREOTYPE_USAGE
        ],
        client_nf_uuid=name_exemplar_nf_uuid,
        client_collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
        stereotype_nf_uuid=digitalisation_level_stereotype_nf_uuid,
    )

    add_new_connector_to_dictionary(
        new_connector_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CONNECTORS
        ],
        place_1_nf_uuid=name_exemplar_nf_uuid,
        place_2_nf_uuid=name_exemplar_type_nf_uuid,
        connector_type=EaConnectorTypes.DEPENDENCY,
    )

    association_nf_uuid = add_new_connector_to_dictionary(
        new_connector_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CONNECTORS
        ],
        place_1_nf_uuid=name_exemplar_nf_uuid,
        place_2_nf_uuid=attributed_object_nf_uuid,
        connector_type=EaConnectorTypes.ASSOCIATION,
    )

    add_new_stereotype_usage_to_dictionary(
        new_stereotype_usage_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.STEREOTYPE_USAGE
        ],
        client_nf_uuid=association_nf_uuid,
        client_collection_type=NfEaComCollectionTypes.EA_CONNECTORS,
        stereotype_nf_uuid=exemplified_by_stereotype_nf_uuid,
    )

    add_new_stereotype_usage_to_dictionary(
        new_stereotype_usage_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.STEREOTYPE_USAGE
        ],
        client_nf_uuid=name_exemplar_nf_uuid,
        client_collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
        stereotype_nf_uuid=name_exemplar_stereotype_nf_uuid,
    )



from enum import auto, unique

from bclearer_core.nf.types.column_types import (
    ColumnTypes,
)


@unique
class NfColumnTypes(ColumnTypes):
    NF_UUIDS = auto()

    UNIVERSE_KEYS = auto()

    COLLECTION_TYPES = auto()

    def __column_name(self) -> str:
        column_name = (
            column_name_mapping[self]
        )

        return column_name

    column_name = property(
        fget=__column_name,
    )


column_name_mapping = {
    NfColumnTypes.NF_UUIDS: "nf_uuids",
    NfColumnTypes.UNIVERSE_KEYS: "universe_keys",
    NfColumnTypes.COLLECTION_TYPES: "collection_types",
}



class NfVerses:
    def __init__(self):
        pass

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from bclearer_core.constants.standard_constants import (
    HR_DATE_TIME_FORMAT,
    HR_NAME_COMPONENT_DIVIDER,
)
from bclearer_core.infrastructure.session.common.session_identity_vectors import (
    OsUserProfileSessionObjectIdentityVector,
    OsUserProfileSessionObjectIdentityVectorPlaces,
)
from bclearer_core.infrastructure.session.common.session_objects import (
    SessionObjects,
)
from bclearer_core.infrastructure.session.operating_system.operating_system_objects import (
    OperatingSystemObjects,
)
from bclearer_core.infrastructure.session.os_user_session.bie_os_user_profile_session_types import (
    BieOsUserProfileSessionTypes,
)
from bclearer_core.infrastructure.session.os_user_session.helpers.session_id_getter import (
    get_os_user_session_id,
)
from bclearer_core.infrastructure.session.os_user_session.helpers.session_start_date_time_getter import (
    get_user_session_start_date_time,
)
from bclearer_core.infrastructure.session.user.os_user_profile_objects import (
    OsUserProfileObjects,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.bie_id_creation_module.bie_id_creation_facade import (
    BieIdCreationFacade,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.objects.bie_ids import (
    BieIds,
)


class OsUserProfileSessionObjects(
    SessionObjects
):
    _instance = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(
                OsUserProfileSessionObjects,
                cls,
            ).__new__(cls)

        return cls._instance

    def __init__(self):
        session_identifier = (
            get_os_user_session_id()
        )

        self.session_id = (
            session_identifier
        )

        self.windows_session_id = (
            session_identifier
        )

        self.os_user_session_id = (
            session_identifier
        )

        self.owning_os_user_profile = (
            OsUserProfileObjects()
        )

        self.owning_operating_system_object = (
            OperatingSystemObjects()
        )

        # TODO: Every time it runs it retrieves a different value (very similar though), but it should retrieve exactly
        #  the same if it's run within the same user session
        self.session_start_date_time = get_user_session_start_date_time(
            user_name=self.owning_os_user_profile.user_name
        )

        # TODO: What enum to use? What design?
        self.os_user_session_type = (
            BieOsUserProfileSessionTypes.BIE_OS_USER_PROFILE_SESSION
        )

        bie_id = self.__create_bie_id()

        base_hr_name = (
            self.owning_operating_system_object.base_hr_name
            + HR_NAME_COMPONENT_DIVIDER
            + self.owning_os_user_profile.base_hr_name
            + HR_NAME_COMPONENT_DIVIDER
            + self.session_start_date_time.strftime(
                format=HR_DATE_TIME_FORMAT
            )
        )

        super().__init__(
            bie_id=bie_id,
            base_hr_name=base_hr_name,
            bie_infrastructure_type=BieOsUserProfileSessionTypes.BIE_OS_USER_PROFILE_SESSION,
        )

    def __create_bie_id(self) -> BieIds:
        identity_vector = OsUserProfileSessionObjectIdentityVector(
            places=OsUserProfileSessionObjectIdentityVectorPlaces(
                os_user_session_type_bie_id=self.os_user_session_type.item_bie_identity,
                windows_session_id=self.windows_session_id,
                session_start_date_time=self.session_start_date_time,
                owning_operating_system_bie_id=self.owning_operating_system_object.bie_id,
            )
        )

        bie_id = BieIdCreationFacade.create_bie_id_from_identity_vector(
            identity_vector=identity_vector,
        )

        return bie_id



from bclearer_core.objects.registries.nf_verse_registries import (
    NfVerseRegistries,
)


class NfUniverseRegistries(
    NfVerseRegistries
):
    def __init__(self):
        super().__init__()

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from bclearer_core.common_knowledge.matched_objects import (
    MatchedEaObjects,
)


class AttributeNameToNameObjectConfigurationObjects:
    def __init__(
        self,
        matched_naming_space_instance: MatchedEaObjects,
    ):
        self.matched_naming_space_instance = matched_naming_space_instance

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from bclearer_core.configurations.bespoke_name_to_instance_configuration_objects import (
    BespokeNameToInstanceConfigurationObjects,
)
from bclearer_core.substages.operations.b_evolve.common.new_root_package_creator import (
    create_root_package,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_names_and_instances.names_and_instances_annotator import (
    annotate_names_and_instances,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_names_and_instances.names_and_instances_separator import (
    separate_names_and_instances,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)


def shift_convention_separate_bespoke_names_and_instances(
    content_universe: NfEaComUniverses,
    list_of_configuration_objects: list,
    output_universe: NfEaComUniverses,
    package_name: str,
) -> NfEaComUniverses:
    log_message(
        message="Separate Bespoke Names and Instances - started",
    )

    output_universe.nf_ea_com_registry.dictionary_of_collections = (
        content_universe.nf_ea_com_registry.dictionary_of_collections.copy()
    )

    for (
        configuration_object
    ) in list_of_configuration_objects:
        __separate_names(
            configuration_object=configuration_object,
            nf_ea_com_universe=output_universe,
            package_name=package_name,
        )

    log_message(
        message="Separate Bespoke Names and Instances - finished",
    )

    return output_universe


def __separate_names(
    configuration_object: BespokeNameToInstanceConfigurationObjects,
    nf_ea_com_universe: NfEaComUniverses,
    package_name: str,
):
    package_nf_uuid = create_root_package(
        nf_ea_com_universe=nf_ea_com_universe,
        package_name=package_name,
    )

    if (
        configuration_object.matched_name_instance_type
        is None
    ):
        separate_names_and_instances(
            nf_ea_com_universe=nf_ea_com_universe,
            matched_naming_space_type=configuration_object.matched_naming_space_type,
            name_instance_attribute_name=configuration_object.name_instance_attribute_name,
            package_nf_uuid=package_nf_uuid,
        )

    else:
        annotate_names_and_instances(
            nf_ea_com_universe=nf_ea_com_universe,
            matched_name_instance_type=configuration_object.matched_name_instance_type,
        )



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.common.ea_guid_from_nf_uuid_creator import (
    create_ea_guid_from_nf_uuid,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.column_types.ea_t.ea_t_package_column_types import (
    EaTPackageColumnTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_element_types import (
    EaElementTypes,
)
from bclearer_orchestration_services.identification_services.uuid_service.uuid_helpers.uuid_factory import (
    create_new_uuid,
)
from pandas import DataFrame, set_option


def create_new_subpackage_if_not_exist(
    nf_ea_com_universe_ea_packages: DataFrame,
    new_ea_packages_dictionary: dict,
    package_nf_uuid: str,
    new_subpackage_name: str,
) -> str:
    nf_uuids_column_name = (
        NfColumnTypes.NF_UUIDS.column_name
    )

    object_name_column_name = (
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name
    )

    contained_ea_packages_column_name = (
        NfEaComColumnTypes.PACKAGES_CONTAINED_EA_PACKAGES.column_name
    )

    list_of_first_level_contained_packages = nf_ea_com_universe_ea_packages[
        nf_ea_com_universe_ea_packages[
            nf_uuids_column_name
        ]
        == package_nf_uuid
    ][
        contained_ea_packages_column_name
    ].tolist()[
        0
    ]

    # Get the first-level contained packages safely
    filtered_packages = nf_ea_com_universe_ea_packages[
        nf_ea_com_universe_ea_packages[
            nf_uuids_column_name
        ]
        == package_nf_uuid
    ]

    if not filtered_packages.empty:
        contained_packages = filtered_packages[
            contained_ea_packages_column_name
        ].tolist()

        # Check if the list is not empty and extract the first element
        if (
            contained_packages
            and contained_packages[0]
            is not None
        ):
            list_of_first_level_contained_packages = contained_packages[
                0
            ]
        else:
            list_of_first_level_contained_packages = (
                []
            )  # Default to an empty list if no contained packages
    else:
        list_of_first_level_contained_packages = (
            []
        )

    # Now proceed with the next steps
    ea_packages_sliced_to_first_level_contained_packages = nf_ea_com_universe_ea_packages[
        nf_ea_com_universe_ea_packages[
            nf_uuids_column_name
        ].isin(
            list_of_first_level_contained_packages
        )
    ]

    # TODO: review this the new code is correct - patched as a result of fixes to objects_to_classes_conventions_shifter.py
    # ea_packages_sliced_to_first_level_contained_packages = nf_ea_com_universe_ea_packages[
    #     nf_ea_com_universe_ea_packages[
    #         nf_uuids_column_name
    #     ].isin(
    #         list_of_first_level_contained_packages,
    #     )
    # ]

    list_of_first_level_contained_packages_names = ea_packages_sliced_to_first_level_contained_packages[
        object_name_column_name
    ].tolist()

    name_package_already_created_in_parent_package_nf_uuid = (
        None
    )

    for (
        new_ea_packages_row_dictionary
    ) in (
        new_ea_packages_dictionary.values()
    ):
        if (
            new_ea_packages_row_dictionary[
                NfEaComColumnTypes.PACKAGEABLE_OBJECTS_PARENT_EA_ELEMENT.column_name
            ]
            == package_nf_uuid
            and new_ea_packages_row_dictionary[
                object_name_column_name
            ]
            == new_subpackage_name
        ):
            name_package_already_created_in_parent_package_nf_uuid = new_ea_packages_row_dictionary[
                nf_uuids_column_name
            ]

    if (
        new_subpackage_name
        in list_of_first_level_contained_packages_names
    ):
        new_subpackage_nf_uuid_row = ea_packages_sliced_to_first_level_contained_packages[
            ea_packages_sliced_to_first_level_contained_packages[
                object_name_column_name
            ]
            == new_subpackage_name
        ]

        new_subpackage_nf_uuid = (
            new_subpackage_nf_uuid_row[
                nf_uuids_column_name
            ]
            .to_string(index=False)
            .strip()
        )

        return new_subpackage_nf_uuid

    if (
        name_package_already_created_in_parent_package_nf_uuid
        is not None
    ):
        return name_package_already_created_in_parent_package_nf_uuid

    new_subpackage_nf_uuid = __add_new_package_row(
        ea_packages_table=nf_ea_com_universe_ea_packages,
        new_ea_packages_dictionary=new_ea_packages_dictionary,
        parent_package_nf_uuid=package_nf_uuid,
        new_package_name=new_subpackage_name,
    )

    list_of_first_level_contained_packages_updated = list_of_first_level_contained_packages.append(
        new_subpackage_nf_uuid,
    )

    nf_ea_com_universe_ea_packages.loc[
        nf_ea_com_universe_ea_packages[
            nf_uuids_column_name
        ]
        == package_nf_uuid,
        contained_ea_packages_column_name,
    ] = list_of_first_level_contained_packages_updated

    # nf_ea_com_universe_ea_packages[
    #     nf_ea_com_universe_ea_packages[
    #         nf_uuids_column_name
    #     ]
    #     == package_nf_uuid
    # ][
    #     contained_ea_packages_column_name
    # ] = list_of_first_level_contained_packages_updated

    return new_subpackage_nf_uuid


def __add_new_package_row(
    ea_packages_table: DataFrame,
    new_ea_packages_dictionary: dict,
    parent_package_nf_uuid: str,
    new_package_name: str,
) -> str:
    nf_uuids_column_name = (
        NfColumnTypes.NF_UUIDS.column_name
    )

    ea_guid_column_name = (
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
    )

    object_name_column_name = (
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name
    )

    new_package_nf_uuid = (
        create_new_uuid()
    )

    new_package_ea_guid = (
        create_ea_guid_from_nf_uuid(
            nf_uuid=new_package_nf_uuid,
        )
    )

    parent_package_ea_guid = (
        ea_packages_table[
            ea_packages_table[
                nf_uuids_column_name
            ]
            == parent_package_nf_uuid
        ][ea_guid_column_name]
        .to_string(index=False)
        .strip()
    )

    parent_package_name = (
        ea_packages_table[
            ea_packages_table[
                nf_uuids_column_name
            ]
            == parent_package_nf_uuid
        ][object_name_column_name]
        .to_string(index=False)
        .strip()
    )

    # Note: Pandas truncates long strings. The following instructions prevents pandas from doing that.
    # TODO: Consider to move this option higher in the hierarchy
    set_option(
        "display.max_colwidth",
        None,
    )

    parent_package_ea_package_path = (
        ea_packages_table[
            ea_packages_table[
                nf_uuids_column_name
            ]
            == parent_package_nf_uuid
        ]["ea_package_path"]
        .to_string(index=False)
        .strip()
    )

    new_package_ea_package_path = (
        parent_package_ea_package_path
        + "/"
        + new_package_name
    )

    ea_packages_row_dictionary = {
        EaTPackageColumnTypes.T_PACKAGE_EA_GUIDS.nf_column_name: new_package_ea_guid,
        "ea_package_path": new_package_ea_package_path,
        "parent_package_ea_guid": parent_package_ea_guid,
        "parent_package_name": parent_package_name,
        nf_uuids_column_name: new_package_nf_uuid,
        NfEaComColumnTypes.PACKAGES_CONTAINED_EA_PACKAGES.column_name: [],
        NfEaComColumnTypes.ELEMENTS_EA_OBJECT_TYPE.column_name: EaElementTypes.PACKAGE.type_name,
        NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name: [],
        NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS.column_name: [],
        NfEaComColumnTypes.ELEMENTS_CONTAINED_EA_DIAGRAMS.column_name: [],
        NfEaComColumnTypes.ELEMENTS_CONTAINED_EA_CLASSIFIERS.column_name: [],
        NfEaComColumnTypes.PACKAGEABLE_OBJECTS_PARENT_EA_ELEMENT.column_name: parent_package_nf_uuid,
        NfEaComColumnTypes.STEREOTYPEABLE_OBJECTS_EA_OBJECT_STEREOTYPES.column_name: [],
        object_name_column_name: new_package_name,
        ea_guid_column_name: new_package_ea_guid,
    }

    new_ea_packages_dictionary[
        len(new_ea_packages_dictionary)
        + 1
    ] = ea_packages_row_dictionary

    return new_package_nf_uuid



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.common.list_of_contained_packages_of_a_package_getter import (
    get_list_of_all_contained_packages_of_package,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_element_types import (
    EaElementTypes,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)
from pandas import DataFrame


# TODO: Use nf ea com table ..._ea_full_packages for this, rather than using recursion
def get_all_package_and_subpackages_classifiers(
    content_universe: NfEaComUniverses,
    package_ea_guid: str,
) -> DataFrame:
    content_universe_ea_packages = (
        content_universe.nf_ea_com_registry.get_ea_packages()
    )

    nf_uuids_column_name = (
        NfColumnTypes.NF_UUIDS.column_name
    )

    list_of_package_ea_guids = content_universe_ea_packages[
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
    ].tolist()

    if (
        package_ea_guid
        not in list_of_package_ea_guids
    ):
        log_message(
            message="CONVENTION SHIFT OPERATION: Extract UML names to attributes - Cannot find package EA GUID: "
            + package_ea_guid,
        )

        return DataFrame()

    package_nf_uuid = (
        content_universe_ea_packages[
            content_universe_ea_packages[
                NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
            ]
            == package_ea_guid
        ][nf_uuids_column_name]
        .to_string(index=False)
        .strip()
    )

    all_packages_and_subpackages_nf_uuids_set = {
        package_nf_uuid,
    }

    all_packages_and_subpackages_nf_uuids_set = get_list_of_all_contained_packages_of_package(
        ea_packages_table=content_universe_ea_packages,
        package_nf_uuid=package_nf_uuid,
        set_of_all_packages_and_subpackages=all_packages_and_subpackages_nf_uuids_set,
    )

    content_universe_ea_classifiers = (
        content_universe.nf_ea_com_registry.get_ea_classifiers()
    )

    parent_ea_element_column_name = (
        NfEaComColumnTypes.PACKAGEABLE_OBJECTS_PARENT_EA_ELEMENT.column_name
    )

    ea_object_type_column_name = (
        NfEaComColumnTypes.ELEMENTS_EA_OBJECT_TYPE.column_name
    )

    domain_objects_classifiers_table = content_universe_ea_classifiers[
        content_universe_ea_classifiers[
            parent_ea_element_column_name
        ].isin(
            all_packages_and_subpackages_nf_uuids_set,
        )
    ]

    domain_classes_classifiers_table = domain_objects_classifiers_table[
        domain_objects_classifiers_table[
            ea_object_type_column_name
        ]
        == EaElementTypes.CLASS.type_name
    ]

    # TODO: Track the export to EA to check where and why the EA GUIDs are being deleted/changed
    # TODO: Reformat instrument_and_visualize to export to EA without deleting/changing the EA GUIDs
    #  - look at AMis test "csv assemblies"

    return (
        domain_classes_classifiers_table
    )



from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)


def create_new_ea_objects_dictionary() -> (
    dict
):
    new_ea_objects_dictionary = {
        NfEaComCollectionTypes.EA_CLASSIFIERS: dict(),
        NfEaComCollectionTypes.EA_CONNECTORS: dict(),
        NfEaComCollectionTypes.EA_PACKAGES: dict(),
        NfEaComCollectionTypes.EA_ATTRIBUTES: dict(),
        NfEaComCollectionTypes.STEREOTYPE_USAGE: dict(),
        NfEaComCollectionTypes.EA_STEREOTYPES: dict(),
    }

    return new_ea_objects_dictionary



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from pandas import DataFrame


def remove_ea_attributes(
    nf_ea_com_universe: NfEaComUniverses,
    ea_attributes: DataFrame,
):
    ea_attribute_nf_uuids_to_remove = set(
        ea_attributes[
            NfColumnTypes.NF_UUIDS.column_name
        ],
    )

    ea_attributes = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_attributes()
    )

    remaining_ea_attributes = ea_attributes[
        ~ea_attributes[
            NfColumnTypes.NF_UUIDS.column_name
        ].isin(
            ea_attribute_nf_uuids_to_remove,
        )
    ]

    nf_ea_com_universe.nf_ea_com_registry.replace_collection(
        collection_type=NfEaComCollectionTypes.EA_ATTRIBUTES,
        collection=remaining_ea_attributes,
    )



from bclearer_core.constants.standard_constants import (
    DEFAULT_NULL_VALUE,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.common.ea_guid_from_nf_uuid_creator import (
    create_ea_guid_from_nf_uuid,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)
from bclearer_interop_services.tuple_service.tuple_attribute_value_getter import (
    get_tuple_attribute_value_if_required,
)
from bclearer_orchestration_services.identification_services.uuid_service.uuid_helpers.uuid_factory import (
    create_new_uuid,
)


def create_dependency_from_attribute(
    attribute_to_convert_tuple: tuple,
    type_nf_uuid: str,
) -> dict:
    dependency_nf_uuid = (
        create_new_uuid()
    )

    dependency_ea_guid = (
        create_ea_guid_from_nf_uuid(
            nf_uuid=dependency_nf_uuid
        )
    )

    dependency_supplier_nf_uuid = (
        type_nf_uuid
    )

    dependency_client_nf_uuid = get_tuple_attribute_value_if_required(
        owning_tuple=attribute_to_convert_tuple,
        attribute_name=NfEaComColumnTypes.ELEMENT_COMPONENTS_CLASSIFYING_EA_CLASSIFIER.column_name,
    )

    ea_dependency_dictionary = {
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name: dependency_ea_guid,
        NfColumnTypes.NF_UUIDS.column_name: dependency_nf_uuid,
        NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name: dependency_supplier_nf_uuid,
        NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS.column_name: dependency_client_nf_uuid,
        NfEaComColumnTypes.CONNECTORS_DIRECTION_TYPE_NAME.column_name: "Source -> Destination",
        NfEaComColumnTypes.CONNECTORS_ELEMENT_TYPE_NAME.column_name: EaConnectorTypes.DEPENDENCY.type_name,
        NfEaComColumnTypes.CONNECTORS_SOURCE_CARDINALITY.column_name: DEFAULT_NULL_VALUE,
        NfEaComColumnTypes.CONNECTORS_DEST_CARDINALITY.column_name: DEFAULT_NULL_VALUE,
        NfEaComColumnTypes.STEREOTYPEABLE_OBJECTS_EA_OBJECT_STEREOTYPES.column_name: [],
        NfEaComColumnTypes.REPOSITORIED_OBJECTS_EA_REPOSITORY.column_name: DEFAULT_NULL_VALUE,
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name: DEFAULT_NULL_VALUE,
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NOTES.column_name: DEFAULT_NULL_VALUE,
    }

    return ea_dependency_dictionary



from bclearer_core.constants.standard_constants import (
    DEFAULT_NULL_VALUE,
)
from bclearer_core.substages.operations.b_evolve.adjustment_operations.convert_attributes_to_associations.content_adjustment_universes_nf_uuids_mapper import (
    get_mapped_nf_uuid_from_mapped_universe,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)
from pandas import DataFrame


def get_attributes_to_convert(
    content_universe: NfEaComUniverses,
    adjustment_universe: NfEaComUniverses,
) -> DataFrame:
    content_collection_dictionary = (
        content_universe.nf_ea_com_registry.dictionary_of_collections
    )

    adjustment_collection_dictionary = (
        adjustment_universe.nf_ea_com_registry.dictionary_of_collections
    )

    adjustment_universe_ea_attributes_table = adjustment_collection_dictionary[
        NfEaComCollectionTypes.EA_ATTRIBUTES
    ]

    filter_columns = [
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name,
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name,
        NfEaComColumnTypes.ATTRIBUTES_LOWER_BOUNDS.column_name,
        NfEaComColumnTypes.ATTRIBUTES_UPPER_BOUNDS.column_name,
    ]

    attribute_conversion_table = adjustment_universe_ea_attributes_table.filter(
        items=filter_columns,
    )

    attribute_conversion_table[
        NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name
    ] = adjustment_universe_ea_attributes_table[
        NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name
    ].apply(
        lambda source_nf_uuid: get_mapped_nf_uuid_from_mapped_universe(
            source_nf_uuid=source_nf_uuid,
            source_universe_collection_dictionary=adjustment_collection_dictionary,
            mapped_universe_collection_dictionary=content_collection_dictionary,
        ),
    )

    # TODO: Find out why the NamedPlace object has BoroDefaultNullValue as attribute type nf_uuid for GeographicalName
    #  in the source data (attribute value: 'name') - and report it
    attribute_conversion_table[
        NfEaComColumnTypes.ELEMENT_COMPONENTS_CLASSIFYING_EA_CLASSIFIER.column_name
    ] = adjustment_universe_ea_attributes_table[
        NfEaComColumnTypes.ELEMENT_COMPONENTS_CLASSIFYING_EA_CLASSIFIER.column_name
    ].apply(
        lambda source_nf_uuid: get_mapped_nf_uuid_from_mapped_universe(
            source_nf_uuid=source_nf_uuid,
            source_universe_collection_dictionary=adjustment_collection_dictionary,
            mapped_universe_collection_dictionary=content_collection_dictionary,
        ),
    )

    cleaned_attribute_conversion_table = __get_cleaned_attribute_conversion_table(
        attribute_conversion_table=attribute_conversion_table,
    )

    return cleaned_attribute_conversion_table


def __get_cleaned_attribute_conversion_table(
    attribute_conversion_table: DataFrame,
) -> DataFrame:
    __log_failed_conversions(
        attribute_conversion_table=attribute_conversion_table,
    )

    cleaned_attribute_conversion_table = attribute_conversion_table.loc[
        attribute_conversion_table[
            NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name
        ]
        != DEFAULT_NULL_VALUE
    ]

    cleaned_attribute_conversion_table = cleaned_attribute_conversion_table.loc[
        cleaned_attribute_conversion_table[
            NfEaComColumnTypes.ELEMENT_COMPONENTS_CLASSIFYING_EA_CLASSIFIER.column_name
        ]
        != DEFAULT_NULL_VALUE
    ]

    return cleaned_attribute_conversion_table


def __log_failed_conversions(
    attribute_conversion_table: DataFrame,
) -> None:
    attributes_with_no_attributed_object = attribute_conversion_table.loc[
        attribute_conversion_table[
            NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name
        ]
        == DEFAULT_NULL_VALUE
    ]

    if (
        not attributes_with_no_attributed_object.empty
    ):
        for (
            bad_attributed_objects_row
        ) in (
            attributes_with_no_attributed_object.iterrows()
        ):
            log_message(
                message="Attribute with name: "
                + str(
                    bad_attributed_objects_row[
                        1
                    ][
                        1
                    ],
                )
                + " - and EA GUID: "
                + str(
                    bad_attributed_objects_row[
                        1
                    ][
                        0
                    ],
                )
                + " - cannot be converted. It has no attributed object.",
            )

    attributes_with_no_attribute_type = attribute_conversion_table.loc[
        attribute_conversion_table[
            NfEaComColumnTypes.ELEMENT_COMPONENTS_CLASSIFYING_EA_CLASSIFIER.column_name
        ]
        == DEFAULT_NULL_VALUE
    ]

    if (
        not attributes_with_no_attribute_type.empty
    ):
        for (
            bad_attribute_types_row
        ) in (
            attributes_with_no_attribute_type.iterrows()
        ):
            log_message(
                message="Attribute with name: "
                + str(
                    bad_attribute_types_row[
                        1
                    ][
                        1
                    ],
                )
                + " - and EA GUID: "
                + str(
                    bad_attribute_types_row[
                        1
                    ][
                        0
                    ],
                )
                + " - cannot be converted. It has no attribute type.",
            )



from bclearer_core.common_knowledge.matched_objects import (
    MatchedEaObjects,
)
from bclearer_core.configurations.universe_modification_configuration_objects import (
    UniverseModificationConfigurationObjects,
)


class AddDependencyToInstancesOfTypeConfigurationObjects(
    UniverseModificationConfigurationObjects,
):
    def __init__(
        self,
        matched_target_type: MatchedEaObjects,
        matched_source_objects_type: MatchedEaObjects,
    ):
        super().__init__()

        self.matched_target_type = (
            matched_target_type
        )

        self.matched_source_objects_type = (
            matched_source_objects_type
        )

    def get_matched_target_type(
        self,
    ) -> MatchedEaObjects:
        return self.matched_target_type

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from bclearer_core.bie.infrastructure.types.bie_common_column_names import (
    BieCommonColumnNames,
)
from bclearer_core.constants.standard_constants import (
    HR_DATE_TIME_FORMAT,
)
from bclearer_core.infrastructure.session.app_run_session.app_run_session_objects import (
    AppRunSessionObjects,
)
from bclearer_core.infrastructure.session.app_run_session.enums.app_run_session_object_column_names import (
    AppRunSessionObjectColumnNames,
)
from bclearer_interop_services.b_dictionary_service.table_as_dictionary_service.table_as_dictionary_to_dataframe_converter import (
    convert_table_as_dictionary_to_dataframe,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.common_knowledge.bie_column_names import (
    BieColumnNames,
)
from pandas import DataFrame


def create_app_run_session_objects_table(
    app_run_session_object: AppRunSessionObjects,
) -> DataFrame:
    table_as_dictionary = {}

    __add_app_run_session_object_to_table(
        table_as_dictionary=table_as_dictionary,
        app_run_session_object=app_run_session_object,
    )

    table_as_dataframe = convert_table_as_dictionary_to_dataframe(
        table_as_dictionary=table_as_dictionary
    )

    return table_as_dataframe.astype(
        str
    )


def __add_app_run_session_object_to_table(
    table_as_dictionary: dict,
    app_run_session_object: AppRunSessionObjects,
) -> None:
    start_date_time = app_run_session_object.app_run_session_date_time.strftime(
        format=HR_DATE_TIME_FORMAT
    )

    app_run_session_object_row = {
        BieColumnNames.BIE_IDS.b_enum_item_name: app_run_session_object.bie_id,
        BieCommonColumnNames.BASE_HR_NAMES.b_enum_item_name: app_run_session_object.base_hr_name,
        BieCommonColumnNames.BIE_INFRASTRUCTURE_TYPE_IDS.b_enum_item_name: app_run_session_object.bie_type.item_bie_identity,
        AppRunSessionObjectColumnNames.START_DATE_TIMES.b_enum_item_name: start_date_time,
        AppRunSessionObjectColumnNames.OWNING_OS_USER_PROFILE_SESSION_BIE_IDS.b_enum_item_name: app_run_session_object.owning_os_user_profile_session.bie_id,
    }

    table_as_dictionary[
        len(table_as_dictionary)
    ] = app_run_session_object_row



"""Enumeration of supported operating system kinds."""

from enum import Enum, auto


# TODO 20251002 - should this be a subclass of BEnums or BieEnums???
class OperatingSystemKind(Enum):
    WINDOWS = auto()
    LINUX = auto()
    MACOS = auto()
    UNKNOWN = auto()



from datetime import datetime
from typing import NamedTuple, Optional

from bclearer_orchestration_services.identification_services.b_identity_ecosystem.bie_id_creation_module.common_knowledge.types.bie_vector_structure_types import (
    BieVectorStructureTypes,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.bie_id_creation_module.identity_vectors.bie_identity_vector_base import (
    BieIdentityVectorBase,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.common_knowledge.bie_enums import (
    BieEnums,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.objects.bie_ids import (
    BieIds,
)


def _validate_places(
    places: NamedTuple,
) -> None:
    if not isinstance(places, tuple):
        raise TypeError(
            "places must be a NamedTuple"
        )

    for field_name in places._fields:
        value = getattr(places, field_name)

        if value is None:
            raise ValueError(
                f"places field '{field_name}' must not be None"
            )


# --- Operating System ---


class OperatingSystemObjectIdentityVectorPlaces(
    NamedTuple,
):
    operating_system_type_bie_id: BieIds
    machine_sid: str


class OperatingSystemObjectIdentityVector(
    BieIdentityVectorBase,
):
    def __init__(
        self,
        places: OperatingSystemObjectIdentityVectorPlaces,
    ):
        _validate_places(places=places)

        self._places = places

    @property
    def bie_domain_type(
        self,
    ) -> Optional[BieEnums]:
        return None

    @property
    def bie_vector_structure_type(
        self,
    ) -> BieVectorStructureTypes:
        return (
            BieVectorStructureTypes.MULTI_DIMENSIONAL_ORDER_SENSITIVE
        )

    def input_objects(self) -> list:
        return [
            self._places.operating_system_type_bie_id,
            self._places.machine_sid,
        ]


# --- OS User Profile ---


class OsUserProfileObjectIdentityVectorPlaces(
    NamedTuple,
):
    os_user_profile_type_bie_id: BieIds
    user_sid: str
    operating_system_type_bie_id: BieIds


class OsUserProfileObjectIdentityVector(
    BieIdentityVectorBase,
):
    def __init__(
        self,
        places: OsUserProfileObjectIdentityVectorPlaces,
    ):
        _validate_places(places=places)

        self._places = places

    @property
    def bie_domain_type(
        self,
    ) -> Optional[BieEnums]:
        return None

    @property
    def bie_vector_structure_type(
        self,
    ) -> BieVectorStructureTypes:
        return (
            BieVectorStructureTypes.MULTI_DIMENSIONAL_ORDER_SENSITIVE
        )

    def input_objects(self) -> list:
        return [
            self._places.os_user_profile_type_bie_id,
            self._places.user_sid,
            self._places.operating_system_type_bie_id,
        ]


# --- OS User Profile Session ---


class OsUserProfileSessionObjectIdentityVectorPlaces(
    NamedTuple,
):
    os_user_session_type_bie_id: BieIds
    windows_session_id: int
    session_start_date_time: datetime
    owning_operating_system_bie_id: BieIds

    @property
    def os_user_session_id(self) -> int:
        return self.windows_session_id


class OsUserProfileSessionObjectIdentityVector(
    BieIdentityVectorBase,
):
    def __init__(
        self,
        places: OsUserProfileSessionObjectIdentityVectorPlaces,
    ):
        _validate_places(places=places)

        self._places = places

    @property
    def bie_domain_type(
        self,
    ) -> Optional[BieEnums]:
        return None

    @property
    def bie_vector_structure_type(
        self,
    ) -> BieVectorStructureTypes:
        return (
            BieVectorStructureTypes.MULTI_DIMENSIONAL_ORDER_SENSITIVE
        )

    def input_objects(self) -> list:
        return [
            self._places.os_user_session_type_bie_id,
            self._places.os_user_session_id,
            self._places.session_start_date_time,
            self._places.owning_operating_system_bie_id,
        ]



"""Windows specific account context helpers."""

from typing import Tuple

from bclearer_core.infrastructure.session.user.helpers.windows.security import (
    lookup_account_sid,
    translate_account_type,
)


def get_windows_account_context(
    user_sid,
) -> Tuple[str, str]:
    (
        account_name,
        domain_name,
        account_type_value,
    ) = lookup_account_sid(user_sid)

    account_type = translate_account_type(
        account_type_value=account_type_value
    )

    return (
        domain_name or account_name,
        account_type,
    )



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.b_evolve.adjustment_operations.convert_attributes_to_associations.association_from_attribute_creator import (
    create_association_from_attribute,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_association_direction_types import (
    EaAssociationDirectionTypes,
)
from bclearer_interop_services.tuple_service.tuple_attribute_value_getter import (
    get_tuple_attribute_value_if_required,
)


def convert_attribute_to_direct_foreign_table_association(
    attribute_to_convert_tuple: tuple,
    new_classifiers_dictionary: dict,
    new_connectors_dictionary: dict,
    direction: EaAssociationDirectionTypes,
) -> tuple:
    new_connectors_dictionary = __add_association(
        attribute_to_convert_tuple=attribute_to_convert_tuple,
        direction=direction,
        new_connectors_dictionary=new_connectors_dictionary,
    )

    return (
        new_classifiers_dictionary,
        new_connectors_dictionary,
    )


def __add_association(
    attribute_to_convert_tuple: tuple,
    direction: EaAssociationDirectionTypes,
    new_connectors_dictionary: dict,
) -> dict:
    type_nf_uuid = get_tuple_attribute_value_if_required(
        owning_tuple=attribute_to_convert_tuple,
        attribute_name=NfEaComColumnTypes.ELEMENT_COMPONENTS_CLASSIFYING_EA_CLASSIFIER.column_name,
    )

    attribute_name = get_tuple_attribute_value_if_required(
        owning_tuple=attribute_to_convert_tuple,
        attribute_name=NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name,
    )

    association_dictionary = create_association_from_attribute(
        attribute_to_convert_tuple=attribute_to_convert_tuple,
        direction=direction,
        type_nf_uuid=type_nf_uuid,
        attribute_name=attribute_name,
    )

    association_nf_uuid = association_dictionary[
        NfColumnTypes.NF_UUIDS.column_name
    ]

    new_connectors_dictionary[
        association_nf_uuid
    ] = association_dictionary

    return new_connectors_dictionary



from bclearer_core.common_knowledge.bclearer_matched_ea_objects import (
    BclearerMatchedEaObjects,
)
from bclearer_core.substages.operations.common.class_adder import (
    add_new_class_to_dictionary,
)
from bclearer_core.substages.operations.common.connector_adder import (
    add_new_connector_to_dictionary,
)
from bclearer_core.substages.operations.common.nf_uuid_from_ea_guid_from_collection_getter import (
    get_nf_uuid_from_ea_guid_from_collection,
)
from bclearer_core.substages.operations.common.stereotype_adder import (
    add_new_stereotype_usage_to_dictionary,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)


def create_name_instance_type(
    nf_ea_com_universe: NfEaComUniverses,
    new_ea_objects_dictionary: dict,
    package_nf_uuid: str,
    name_object_name: str,
    naming_space_nf_uuid: str,
    digitalisation_level_stereotype_nf_uuid: str,
) -> str:
    name_instance_type_nf_uuid = add_new_class_to_dictionary(
        new_classifier_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CLASSIFIERS
        ],
        package_nf_uuid=package_nf_uuid,
        class_name=name_object_name,
    )

    add_new_stereotype_usage_to_dictionary(
        new_stereotype_usage_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.STEREOTYPE_USAGE
        ],
        client_nf_uuid=name_instance_type_nf_uuid,
        client_collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
        stereotype_nf_uuid=digitalisation_level_stereotype_nf_uuid,
    )

    name_instances_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
        ea_guid=BclearerMatchedEaObjects.NAME_INSTANCES.ea_guid,
    )

    add_new_connector_to_dictionary(
        new_connector_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CONNECTORS
        ],
        place_1_nf_uuid=name_instance_type_nf_uuid,
        place_2_nf_uuid=name_instances_nf_uuid,
        connector_type=EaConnectorTypes.GENERALIZATION,
    )

    add_new_connector_to_dictionary(
        new_connector_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CONNECTORS
        ],
        place_1_nf_uuid=name_instance_type_nf_uuid,
        place_2_nf_uuid=naming_space_nf_uuid,
        connector_type=EaConnectorTypes.ASSOCIATION,
        connector_name=BclearerMatchedEaObjects.NAME_TYPES_INSTANCES_STEREOTYPE.object_name,
    )

    return name_instance_type_nf_uuid



from bclearer_core.bie.infrastructure.types.bie_common_column_names import (
    BieCommonColumnNames,
)
from bclearer_core.constants.standard_constants import (
    HR_DATE_TIME_FORMAT,
)
from bclearer_core.infrastructure.session.os_user_session.enums.os_user_profile_session_object_column_names import (
    OsUserProfileSessionObjectColumnNames,
)
from bclearer_core.infrastructure.session.os_user_session.os_user_profile_session_objects import (
    OsUserProfileSessionObjects,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.common_knowledge.bie_column_names import (
    BieColumnNames,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.infrastructure.registrations.bie_infrastructure_registries import (
    BieInfrastructureRegistries,
)


def register_os_user_profile_session_objects(
    bie_infrastructure_registry: BieInfrastructureRegistries,
) -> None:
    os_user_profile_session_object = (
        OsUserProfileSessionObjects()
    )

    bie_infrastructure_registry.register_bie_object_and_type_instance(
        bie_object=os_user_profile_session_object
    )

    bie_infrastructure_registry.register_bie_whole_part_in_bie_infrastructure_registry(
        bie_object_whole=os_user_profile_session_object.owning_operating_system_object,
        bie_object_part=os_user_profile_session_object,
    )

    bie_infrastructure_registry.register_bie_whole_part_in_bie_infrastructure_registry(
        bie_object_whole=os_user_profile_session_object.owning_os_user_profile,
        bie_object_part=os_user_profile_session_object,
    )


def __get_os_user_profile_session_object_table_as_dictionary() -> (
    dict
):
    os_user_profile_session_object = (
        OsUserProfileSessionObjects()
    )

    session_start_date_time = os_user_profile_session_object.session_start_date_time.strftime(
        format=HR_DATE_TIME_FORMAT
    )

    os_user_profile_session_object_row = {
        BieColumnNames.BIE_IDS.b_enum_item_name: os_user_profile_session_object.bie_id,
        BieCommonColumnNames.BASE_HR_NAMES.b_enum_item_name: os_user_profile_session_object.base_hr_name,
        BieCommonColumnNames.BIE_INFRASTRUCTURE_TYPE_IDS.b_enum_item_name: os_user_profile_session_object.bie_type.item_bie_identity,
        OsUserProfileSessionObjectColumnNames.WINDOWS_SESSION_IDS.b_enum_item_name: os_user_profile_session_object.windows_session_id,
        OsUserProfileSessionObjectColumnNames.SESSION_START_DATE_TIMES.b_enum_item_name: session_start_date_time,
        OsUserProfileSessionObjectColumnNames.OWNING_OS_USER_PROFILE_BIE_IDS.b_enum_item_name: os_user_profile_session_object.owning_os_user_profile.bie_id,
        OsUserProfileSessionObjectColumnNames.OWNING_OPERATING_SYSTEM_BIE_IDS.b_enum_item_name: os_user_profile_session_object.owning_operating_system_object.bie_id,
    }

    table_as_dictionary = {
        0: os_user_profile_session_object_row
    }

    return table_as_dictionary



from bclearer_core.infrastructure.session.app_run_session.app_run_session_objects import (
    AppRunSessionObjects,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.infrastructure.registrations.bie_infrastructure_registries import (
    BieInfrastructureRegistries,
)


def register_app_run_session_objects(
    bie_infrastructure_registry: BieInfrastructureRegistries,
    app_run_session_object: AppRunSessionObjects,
) -> None:
    bie_infrastructure_registry.register_bie_object_and_type_instance(
        bie_object=app_run_session_object
    )

    bie_infrastructure_registry.register_bie_whole_part_in_bie_infrastructure_registry(
        bie_object_whole=app_run_session_object.owning_os_user_profile_session,
        bie_object_part=app_run_session_object,
    )



"""Platform aware user session identifier helpers."""

from bclearer_core.infrastructure.session.common.enums.operating_system_kind import (
    OperatingSystemKind,
)
from bclearer_core.infrastructure.session.common.platform_type_getter import (
    get_operating_system_kind,
)
from bclearer_core.infrastructure.session.os_user_session.helpers.unix.unix_session_identifier import (
    get_unix_session_id,
)


def get_os_user_session_id() -> int:
    operating_system_kind = (
        get_operating_system_kind()
    )

    if (
        operating_system_kind
        is OperatingSystemKind.WINDOWS
    ):
        from bclearer_core.infrastructure.session.os_user_session.helpers.windows.windows_user_session_id_getter import (
            get_windows_user_session_id,
        )

        return (
            get_windows_user_session_id()
        )

    return get_unix_session_id()



from bclearer_core.configurations.b_configurations.b_configurations import (
    BConfigurations,
)
from bclearer_orchestration_services.log_environment_utility_service.common_knowledge.environment_log_level_types import (
    EnvironmentLogLevelTypes,
)


class RunEnvironmentLoggingConfigurations(
    BConfigurations
):
    # TODO: added here to be agreed
    ENVIRONMENT_LOG_LEVEL_TYPE = (
        EnvironmentLogLevelTypes.FILTERED
    )



from bclearer_core.common_knowledge.content_operation_types import (
    ContentOperationTypes,
)
from bclearer_core.configurations.content_operation_configurations import (
    ContentOperationConfigurations,
)
from bclearer_core.substages.operations.b_evolve.content_operations.merge_universes.universe_merger import (
    merge_universes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.session.orchestrators.ea_tools_session_managers import (
    EaToolsSessionManagers,
)
from bclearer_interop_services.ea_interop_service.session.processes.creators.empty_nf_ea_com_universe_creator import (
    create_empty_nf_ea_universe,
)


class ContentOperationsSubstages:
    def __init__(
        self,
        ea_tools_session_manager: EaToolsSessionManagers,
        content_operation_configuration: ContentOperationConfigurations,
        content_1_universe: NfEaComUniverses,
        content_2_universe: NfEaComUniverses,
    ):
        self.ea_tools_session_manager = (
            ea_tools_session_manager
        )

        self.content_operation_configuration = content_operation_configuration

        self.content_1_universe = (
            content_1_universe
        )

        self.content_2_universe = (
            content_2_universe
        )

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass

    def run(self) -> NfEaComUniverses:
        output_universe = create_empty_nf_ea_universe(
            ea_tools_session_manager=self.ea_tools_session_manager,
            short_name=self.content_operation_configuration.output_universe_short_name,
        )

        if (
            self.content_operation_configuration.operation_type
            == ContentOperationTypes.MERGE_UNIVERSES
        ):
            merge_universes(
                content_1_universe=self.content_1_universe,
                content_2_universe=self.content_2_universe,
                output_universe=output_universe,
                default_digitalisation_level_stereotype=self.content_operation_configuration.default_digitalisation_level_stereotype,
                context=self.__class__.__name__,
            )

        return output_universe



from abc import ABC
from enum import Enum
from typing import Any, Dict

from bclearer_core.objects.registers.registers import Registers


class Registries(ABC):
    """
    Base class for registry objects that manage collections of registers.
    
    A Registry acts as a container for multiple Registers, typically grouping
    related data sources (e.g., SAP registry contains workorders, notifications, etc.).
    
    A registry can contain registers of different types (e.g., PandasTableRegisters,
    GraphDatabaseRegisters, etc.) as long as they inherit from the Registers base class.
        
    Attributes:
        owning_universe: Back-reference to the parent Universe object.
        _registers: Dictionary mapping enum keys to register instances.
    """
    
    def __init__(
            self,
            owning_universe):
        super().__init__()

        self.owning_universe = \
            owning_universe
        
        self._registers: Dict[Enum, Registers] = \
            {}
        
    @property
    def registers(
            self) \
            -> Dict[Enum, Registers]:
        """
        Get all registers in this registry.
        
        :returns: Dictionary mapping enum keys to register instances.
        """
        return \
            self._registers
    
    def get_register(
            self,
            name: Enum) \
            -> Registers:
        """
        Retrieve a register by its enum name.
        
        :param name: The enum key for the register.
        :returns: The register instance.
        :raises ValueError: If register not found.
        """
        if name not in self._registers:
            raise \
                ValueError(
                    f"Register '{name}' not found in {self.__class__.__name__}")
        
        return \
            self._registers[name]
    
    def set_register(
            self,
            name: Enum,
            register: Registers) \
            -> None:
        """
        Store a register by its enum name.
        
        :param name: The enum key for the register.
        :param register: The register instance to store (can be any Registers subclass).
        """
        self._registers[name] = \
            register
        
        # Also set as attribute for direct access (e.g., registry.workorders)
        setattr(
            self,
            name.value.lower(),
            register)
    
    def has_register(
            self,
            name: Enum) \
            -> bool:
        """
        Check if a register exists by name.
        
        :param name: The enum key to check.
        :returns: True if register exists, False otherwise.
        """
        return \
            name in self._registers
    
    def get_all_register_names(
            self) \
            -> list[Enum]:
        """
        Get a list of all register names in this registry.
        
        :returns: List of enum keys for all registers.
        """
        return \
            list(self._registers.keys())
    
    def set_register_content(
            self,
            register_name: Enum,
            content: Any) \
            -> None:
        """
        Load content into a register by name.
        
        This is a convenience method that gets the register and sets its content.
        
        :param register_name: The enum key for the register.
        :param content: The content to load into the register.
        :raises ValueError: If register not found.
        """
        register = \
            self.get_register(
                name=register_name)
        
        register.set_content(
            content=content)
    
    def get_register_content(
            self,
            register_name: Enum) \
            -> Any:
        """
        Retrieve content from a register by name.
        
        This is a convenience method that gets the register and returns its content.
        
        :param register_name: The enum key for the register.
        :returns: The content stored in the register.
        :raises ValueError: If register not found or has no content.
        """
        register = \
            self.get_register(
                name=register_name)
        
        return \
            register.get_content()
    
    def export_to_disk(
            self,
            folder_path: str) \
            -> None:
        """
        Export all registers to disk.
        
        Default implementation delegates to each register's export_to_disk method.
        Override this method if custom export logic is needed.
        
        :param folder_path: The base folder path for exports.
        """
        for register in self._registers.values():
            register.export_to_disk(
                base_folder_path=folder_path)
    
    def log_registry_state(
            self) \
            -> None:
        """
        Log the state of all registers in this registry.
        
        This method can be overridden for custom logging behavior,
        but the default implementation should work for most cases.
        """
        from bclearer_core.configurations.datastructure.logging_inspection_level_b_enums import \
            LoggingInspectionLevelBEnums
        from bclearer_orchestration_services.reporting_service.reporters.inspection_message_logger import \
            log_inspection_message
        
        registry_name = \
            self.__class__.__name__
        
        log_inspection_message(
            logging_inspection_level_b_enum=LoggingInspectionLevelBEnums.INFO,
            message=f"Registry '{registry_name}' contains {len(self._registers)} register(s).")
        
        for register_name, register in self._registers.items():
            has_content = \
                register.has_content()
            
            log_inspection_message(
                logging_inspection_level_b_enum=LoggingInspectionLevelBEnums.INFO,
                message=f"  - Register '{register_name.value}': {'Has content' if has_content else 'Empty'}")



from pathlib import Path

from bclearer_core.configurations.workspaces.workspace_configurations import (
    WorkspaceConfigurations,
)
from bclearer_interop_services.file_system_service.objects.folders import (
    Folders,
)
from bclearer_orchestration_services.datetime_service.time_helpers.time_getter import (
    now_time_as_string_for_files,
)


def create_workspace_configuration(
    parent_workspace: Folders,
    workspace_type_prefix: (
        str | None
    ) = None,
    workspace_type_suffix: (
        str | None
    ) = None,
) -> WorkspaceConfigurations:
    this_run_human_readable_short_name = (
        now_time_as_string_for_files()
    )

    this_run_workspace = __get_this_run_workspace(
        parent_workspace=parent_workspace,
        this_run_human_readable_short_name=this_run_human_readable_short_name,
        workspace_type_prefix=workspace_type_prefix,
        workspace_type_suffix=workspace_type_suffix,
    )

    workspace_configuration = WorkspaceConfigurations(
        this_run_workspace=this_run_workspace,
        this_run_human_readable_short_name=this_run_human_readable_short_name,
    )

    return workspace_configuration


def __get_this_run_workspace(
    parent_workspace: Folders,
    this_run_human_readable_short_name: str,
    workspace_type_prefix: (
        str | None
    ) = None,
    workspace_type_suffix: (
        str | None
    ) = None,
) -> Folders:
    this_run_workspace_name = __get_this_run_workspace_name(
        this_run_human_readable_short_name=this_run_human_readable_short_name,
        workspace_type_prefix=workspace_type_prefix,
        workspace_type_suffix=workspace_type_suffix,
    )

    this_run_workspace = parent_workspace.get_descendant_file_system_folder(
        relative_path=Path(
            this_run_workspace_name
        )
    )

    this_run_workspace.make_me_on_disk()

    return this_run_workspace


def __get_this_run_workspace_name(
    this_run_human_readable_short_name: str,
    workspace_type_prefix: (
        str | None
    ) = None,
    workspace_type_suffix: (
        str | None
    ) = None,
) -> str:
    workspace_name_components = [
        component
        for component in (
            workspace_type_prefix,
            this_run_human_readable_short_name,
            workspace_type_suffix,
        )
        if component
    ]

    this_run_workspace_name = "_".join(
        workspace_name_components
    )

    return this_run_workspace_name



from bclearer_core.common_knowledge.digitialisation_level_stereotype_matched_ea_objects import (
    DigitalisationLevelStereotypeMatchedEaObjects,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.b_evolve.common.digitalisation_level_stereotypes_usages_getter import (
    get_digitalisation_level_stereotypes_usages,
)
from bclearer_core.substages.operations.b_evolve.common.universes_merge_registers import (
    UniversesMergeRegisters,
)
from bclearer_core.substages.operations.b_evolve.content_operations.merge_universes.nf_ea_com_collection_processes.universes_concatenator import (
    concat_universe_collections,
)
from bclearer_core.substages.operations.common.new_ea_objects_dictionary_creator import (
    create_new_ea_objects_dictionary,
)
from bclearer_core.substages.operations.common.nf_ea_com_universe_updater import (
    update_nf_ea_com_universe_with_dictionary,
)
from bclearer_core.substages.operations.common.nf_uuid_from_ea_guid_from_collection_getter import (
    get_nf_uuid_from_ea_guid_from_collection,
)
from bclearer_core.substages.operations.common.stereotype_adder import (
    add_new_stereotype_usage_to_dictionary,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)


def merge_universes(
    content_1_universe: NfEaComUniverses,
    content_2_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
    default_digitalisation_level_stereotype: DigitalisationLevelStereotypeMatchedEaObjects,
    context="",
) -> NfEaComUniverses:
    __run_input_checks()

    output_universe = __run_operation(
        content_1_universe=content_1_universe,
        content_2_universe=content_2_universe,
        output_universe=output_universe,
        default_digitalisation_level_stereotype=default_digitalisation_level_stereotype,
        context=context,
    )

    return output_universe


def __run_input_checks():
    pass


def __run_operation(
    content_1_universe: NfEaComUniverses,
    content_2_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
    default_digitalisation_level_stereotype: DigitalisationLevelStereotypeMatchedEaObjects,
    context: str,
) -> NfEaComUniverses:
    log_message(
        message="CONTENT OPERATION: Merge universes - "
        + content_1_universe.ea_repository.short_name
        + " + "
        + content_2_universe.ea_repository.short_name
        + " - started",
    )

    universes_merge_register = UniversesMergeRegisters(
        universe_1=content_1_universe,
        universe_2=content_2_universe,
        context=context,
    )

    (
        common_collection_types,
        universe_1_surplus_collection_types,
        universe_2_surplus_collection_types,
    ) = __get_collection_types_distribution_over_universes(
        content_1_universe=content_1_universe,
        content_2_universe=content_2_universe,
    )

    for (
        collection_type
    ) in common_collection_types:
        concat_universe_collections(
            collection_type=collection_type,
            universe_merge_register=universes_merge_register,
            output_universe=output_universe,
        )

    __add_collections(
        input_universe=content_1_universe,
        output_universe=output_universe,
        collection_types=universe_1_surplus_collection_types,
    )

    __add_collections(
        input_universe=content_2_universe,
        output_universe=output_universe,
        collection_types=universe_2_surplus_collection_types,
    )

    if default_digitalisation_level_stereotype:
        __add_default_digitalisation_level_stereotypes(
            nf_ea_com_universe=output_universe,
            default_digitalisation_level_stereotype=default_digitalisation_level_stereotype,
        )

    log_message(
        message="CONTENT OPERATION: Merge universes - "
        + content_1_universe.ea_repository.short_name
        + " + "
        + content_2_universe.ea_repository.short_name
        + " - finished",
    )

    return output_universe


def __get_collection_types_distribution_over_universes(
    content_1_universe: NfEaComUniverses,
    content_2_universe: NfEaComUniverses,
) -> tuple:
    set_of_content_1_collection_types = set(
        content_1_universe.nf_ea_com_registry.dictionary_of_collections.keys(),
    )

    set_of_content_2_collection_types = set(
        content_2_universe.nf_ea_com_registry.dictionary_of_collections.keys(),
    )

    common_collection_types = set_of_content_1_collection_types.intersection(
        set_of_content_2_collection_types,
    )

    universe_1_surplus_collection_types = set_of_content_1_collection_types.difference(
        set_of_content_2_collection_types,
    )

    universe_2_surplus_collection_types = set_of_content_2_collection_types.difference(
        set_of_content_1_collection_types,
    )

    return (
        common_collection_types,
        universe_1_surplus_collection_types,
        universe_2_surplus_collection_types,
    )


def __add_collections(
    input_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
    collection_types: set,
):
    for (
        collection_type
    ) in collection_types:
        output_universe.nf_ea_com_registry.dictionary_of_collections[
            collection_type
        ] = input_universe.nf_ea_com_registry.dictionary_of_collections[
            collection_type
        ]


def __add_default_digitalisation_level_stereotypes(
    nf_ea_com_universe: NfEaComUniverses,
    default_digitalisation_level_stereotype: DigitalisationLevelStereotypeMatchedEaObjects,
) -> None:
    new_ea_objects_dictionary = (
        create_new_ea_objects_dictionary()
    )

    default_digitalisation_level_stereotype_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
        ea_guid=default_digitalisation_level_stereotype.ea_guid,
    )

    digitalisation_level_stereotypes_usages = get_digitalisation_level_stereotypes_usages(
        nf_ea_com_universe=nf_ea_com_universe,
    )

    classifier_digitalisation_level_stereotyped_nf_uuids = set(
        digitalisation_level_stereotypes_usages[
            NfEaComColumnTypes.STEREOTYPE_CLIENT_NF_UUIDS.column_name
        ],
    )

    ea_classifiers = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_classifiers()
    )

    ea_classifiers_without_digitalisation_levels = ea_classifiers[
        ~ea_classifiers[
            NfColumnTypes.NF_UUIDS.column_name
        ].isin(
            classifier_digitalisation_level_stereotyped_nf_uuids,
        )
    ]

    ea_classifiers_without_digitalisation_level_nf_uuids = set(
        ea_classifiers_without_digitalisation_levels[
            NfColumnTypes.NF_UUIDS.column_name
        ],
    )

    for (
        ea_classifier_nf_uuid
    ) in ea_classifiers_without_digitalisation_level_nf_uuids:
        add_new_stereotype_usage_to_dictionary(
            new_stereotype_usage_dictionary=new_ea_objects_dictionary[
                NfEaComCollectionTypes.STEREOTYPE_USAGE
            ],
            client_nf_uuid=ea_classifier_nf_uuid,
            client_collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
            stereotype_nf_uuid=default_digitalisation_level_stereotype_nf_uuid,
        )

    update_nf_ea_com_universe_with_dictionary(
        nf_ea_com_universe=nf_ea_com_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
    )



from bclearer_core.configurations.workspaces.factories.workspace_configuration_factory import (
    create_workspace_configuration,
)
from bclearer_core.configurations.workspaces.workspace_configurations import (
    WorkspaceConfigurations,
)
from bclearer_interop_services.file_system_service.objects.folders import (
    Folders,
)


def create_service_workspace_configuration(
    this_service_workspace: Folders,
    this_service_workspace_type_prefix: (
        str | None
    ) = None,
    this_service_workspace_type_suffix: (
        str | None
    ) = None,
) -> WorkspaceConfigurations:
    service_workspace_configuration = create_workspace_configuration(
        parent_workspace=this_service_workspace,
        workspace_type_prefix=this_service_workspace_type_prefix,
        workspace_type_suffix=this_service_workspace_type_suffix,
    )

    return (
        service_workspace_configuration
    )



"""Windows specific helpers related to user security information."""

import ctypes

import win32security
from bclearer_core.infrastructure.session.common.session_config import (
    UNIX_ADMIN_ACCOUNT_TYPE,
    UNIX_STANDARD_ACCOUNT_TYPE,
)


# TODO: Only have one public method per py file
def get_windows_user_sid():
    token = win32security.OpenProcessToken(
        ctypes.windll.kernel32.GetCurrentProcess(),
        win32security.TOKEN_QUERY,
    )

    user_info = win32security.GetTokenInformation(
        token, win32security.TokenUser
    )

    return user_info[0]


# TODO: where should this be?
def lookup_account_sid(user_sid):
    return (
        win32security.LookupAccountSid(
            None, user_sid
        )
    )


# TODO: should UNIX be in this
def translate_account_type(
    account_type_value,
):
    return (
        UNIX_ADMIN_ACCOUNT_TYPE
        if account_type_value
        == win32security.SidTypeUser
        else UNIX_STANDARD_ACCOUNT_TYPE
    )



from bclearer_core.substages.operations.b_evolve.content_operations.merge_universes.nf_uuid_mapping_processes.uuid_columns_lists_constants import (
    LIST_OF_NF_UUID_COLUMN_NAMES,
    LIST_OF_NF_UUID_LISTS_COLUMN_NAMES,
)
from pandas import DataFrame


def replace_nf_uuids_in_collection(
    column_name: str,
    nf_uuid_mapping_dictionary: dict,
    collection: DataFrame,
) -> DataFrame:
    column_has_lists = (
        __check_column_has_lists(
            column_name=column_name,
            dataframe=collection,
        )
    )

    if column_has_lists:
        if (
            column_name
            in LIST_OF_NF_UUID_LISTS_COLUMN_NAMES
        ):
            collection_with_nf_uuids_replaced = __update_nf_uuids_list_column_in_dataframe(
                nf_uuid_mapping_dictionary=nf_uuid_mapping_dictionary,
                dataframe=collection,
                uuid_list_column_name=column_name,
            )

        else:
            collection_with_nf_uuids_replaced = (
                collection
            )

    elif (
        column_name
        in LIST_OF_NF_UUID_COLUMN_NAMES
    ):
        collection_with_nf_uuids_replaced = __update_nf_uuids_column_in_dataframe(
            nf_uuid_mapping_dictionary=nf_uuid_mapping_dictionary,
            dataframe=collection,
            uuid_column_name=column_name,
        )

    else:
        collection_with_nf_uuids_replaced = (
            collection
        )

    return collection_with_nf_uuids_replaced


def __check_column_has_lists(
    column_name: str,
    dataframe: DataFrame,
) -> bool:
    dataframe_of_types = (
        # TODO: fixed
        dataframe.map(type)
    )

    dataframe_of_lists = (
        dataframe_of_types.loc[
            dataframe_of_types[
                column_name
            ].astype(str)
            == "<class 'list'>"
        ]
    )

    column_has_lists = (
        dataframe_of_lists.shape[0] > 0
    )

    return column_has_lists


def __update_nf_uuids_column_in_dataframe(
    nf_uuid_mapping_dictionary: dict,
    dataframe: DataFrame,
    uuid_column_name: str,
) -> DataFrame:
    updated_dataframe = dataframe.copy()

    updated_dataframe[
        uuid_column_name
    ] = updated_dataframe[
        uuid_column_name
    ].apply(
        lambda original_nf_uuid: __replace_original_nf_uuid_by_new_nf_uuid(
            original_nf_uuid=original_nf_uuid,
            nf_uuid_mapping_dictionary=nf_uuid_mapping_dictionary,
        ),
    )

    return updated_dataframe


def __update_nf_uuids_list_column_in_dataframe(
    nf_uuid_mapping_dictionary: dict,
    dataframe: DataFrame,
    uuid_list_column_name: str,
) -> DataFrame:
    updated_dataframe = dataframe.copy()

    updated_dataframe[
        uuid_list_column_name
    ] = dataframe[
        uuid_list_column_name
    ].apply(
        lambda list_of_uuids: __replace_list_of_uuids(
            list_of_uuids=list_of_uuids,
            nf_uuid_mapping_dictionary=nf_uuid_mapping_dictionary,
        ),
    )

    return updated_dataframe


def __replace_list_of_uuids(
    list_of_uuids: list,
    nf_uuid_mapping_dictionary: dict,
) -> list:
    list_of_new_uuids = []

    if not isinstance(
        list_of_uuids,
        list,
    ):
        return list_of_new_uuids

    for original_uuid in list_of_uuids:
        list_of_new_uuids.append(
            __replace_original_nf_uuid_by_new_nf_uuid(
                original_nf_uuid=original_uuid,
                nf_uuid_mapping_dictionary=nf_uuid_mapping_dictionary,
            ),
        )

    return list_of_new_uuids


def __replace_original_nf_uuid_by_new_nf_uuid(
    original_nf_uuid: str,
    nf_uuid_mapping_dictionary: dict,
) -> str:
    try:
        new_nf_uuid = (
            nf_uuid_mapping_dictionary[
                original_nf_uuid
            ]
        )

    except KeyError:
        new_nf_uuid = original_nf_uuid

    return new_nf_uuid



from bclearer_core.infrastructure.session.app_run_session.app_run_session_objects import (
    AppRunSessionObjects,
)
from bclearer_core.infrastructure.session.dataframes_dictionary.table_creators.app_run_session_objects_table_creator import (
    create_app_run_session_objects_table,
)
from bclearer_core.infrastructure.session.dataframes_dictionary.table_creators.operating_system_objects_table_creator import (
    create_operating_system_objects_table,
)
from bclearer_core.infrastructure.session.dataframes_dictionary.table_creators.os_user_profile_objects_table_creator import (
    create_os_user_profile_objects_table,
)
from bclearer_core.infrastructure.session.dataframes_dictionary.table_creators.os_user_profile_session_objects_table_creator import (
    create_os_user_profile_session_objects_table,
)
from bclearer_core.infrastructure.session.os_user_session.enums.session_object_column_names import (
    SessionObjectColumnNames,
)


def create_session_dataframes_dictionary(
    app_run_session_object: AppRunSessionObjects,
) -> dict:
    app_run_sessions_table = create_app_run_session_objects_table(
        app_run_session_object=app_run_session_object
    )

    operating_system_objects_table = (
        create_operating_system_objects_table()
    )

    os_user_profile_objects_table = (
        create_os_user_profile_objects_table()
    )

    os_user_profile_session_objects_table = (
        create_os_user_profile_session_objects_table()
    )

    dataframes_dictionary_keyed_on_string = {
        SessionObjectColumnNames.APP_RUN_SESSIONS.b_enum_item_name: app_run_sessions_table,
        SessionObjectColumnNames.OPERATING_SYSTEMS.b_enum_item_name: operating_system_objects_table,
        SessionObjectColumnNames.OS_USER_PROFILES.b_enum_item_name: os_user_profile_objects_table,
        SessionObjectColumnNames.OS_USER_PROFILE_SESSIONS.b_enum_item_name: os_user_profile_session_objects_table,
    }

    return dataframes_dictionary_keyed_on_string



"""Template for b_unit class."""

B_UNIT_TEMPLATE = """from bclearer_core.configurations.datastructure.logging_inspection_level_b_enums import (
    LoggingInspectionLevelBEnums,
)
from bclearer_core.objects.b_units import BUnits
from bclearer_orchestration_services.reporting_service.reporters.inspection_message_logger import (
    log_inspection_message,
)


class {class_name}BUnits(BUnits):
    def __init__(self):
        pass

    def run(self) -> None:
        log_inspection_message(
            message="Running bUnit: {{}}".format(
                self.__class__.__name__
            ),
            logging_inspection_level_b_enum=LoggingInspectionLevelBEnums.INFO,
        )

        self.b_unit_process_function()

    def b_unit_process_function(
        self,
    ) -> None:
        pass
"""



from bclearer_core.infrastructure.session.operating_system.operating_system_objects import (
    OperatingSystemObjects,
)
from bclearer_orchestration_services.identification_services.b_identity_ecosystem.infrastructure.registrations.bie_infrastructure_registries import (
    BieInfrastructureRegistries,
)


def register_operating_system_objects(
    bie_infrastructure_registry: BieInfrastructureRegistries,
) -> None:
    operating_system_object = (
        OperatingSystemObjects()
    )

    bie_infrastructure_registry.register_bie_object_and_type_instance(
        bie_object=operating_system_object
    )



import re
from enum import Enum


class BEnums(Enum):
    @classmethod
    def b_enum_name(cls):
        class_name = cls.__qualname__

        pattern = re.compile(
            r"(?<!^)(?=[A-Z])"
        )

        b_enum_name = pattern.sub(
            "_", class_name
        ).lower()

        return b_enum_name

    @property
    def b_enum_item_name(self):
        b_enum_item_name = str.casefold(
            self.name
        )

        return b_enum_item_name



"""macOS specific machine identifier helpers."""

import re
import subprocess
import uuid

from bclearer_core.infrastructure.session.common.session_config import (
    MACOS_IOREG_COMMAND,
    MACOS_IOREG_UUID_PATTERN,
    MACOS_MACHINE_FALLBACK_PREFIX,
)


def get_macos_machine_identifier() -> (
    str
):
    try:
        output = (
            subprocess.check_output(
                MACOS_IOREG_COMMAND,
                text=True,
            )
        )
    except (
        FileNotFoundError,
        subprocess.CalledProcessError,
    ):
        return f"{MACOS_MACHINE_FALLBACK_PREFIX}{uuid.getnode():012x}"

    match = re.search(
        MACOS_IOREG_UUID_PATTERN, output
    )
    if match:
        return match.group(1)

    return f"{MACOS_MACHINE_FALLBACK_PREFIX}{uuid.getnode():012x}"



from dataclasses import dataclass

from bclearer_interop_services.file_system_service.objects.folders import (
    Folders,
)


@dataclass(frozen=True)
class WorkspaceConfigurations:
    this_run_workspace: Folders

    this_run_human_readable_short_name: (
        str
    )



from bclearer_core.configurations.generalise_names_configuration_objects import (
    GeneraliseNamesConfigurationObjects,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.generalise_names.stereotype_instances_associations_adder import (
    add_stereotype_to_instances_associations,
)
from bclearer_core.substages.operations.common.connector_adder import (
    add_new_connector_to_dictionary,
)
from bclearer_core.substages.operations.common.connector_rename import (
    rename_connector,
)
from bclearer_core.substages.operations.common.intersection_getter import (
    get_intersection_of_dependency_and_association_linked,
)
from bclearer_core.substages.operations.common.new_ea_objects_dictionary_creator import (
    create_new_ea_objects_dictionary,
)
from bclearer_core.substages.operations.common.nf_ea_com_universe_updater import (
    update_nf_ea_com_universe_with_dictionary,
)
from bclearer_core.substages.operations.common.nf_uuid_from_ea_guid_from_collection_getter import (
    get_nf_uuid_from_ea_guid_from_collection,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)


def shift_convention_generalise_names(
    content_universe: NfEaComUniverses,
    list_of_configuration_objects: list,
    output_universe: NfEaComUniverses,
) -> NfEaComUniverses:
    log_message(
        message="Generalise Names - started",
    )

    output_universe.nf_ea_com_registry.dictionary_of_collections = (
        content_universe.nf_ea_com_registry.dictionary_of_collections.copy()
    )

    __generalise_naming_spaces(
        list_of_configuration_objects=list_of_configuration_objects,
        nf_ea_com_universe=output_universe,
    )

    log_message(
        message="Generalise Names - finished",
    )

    return output_universe


def __generalise_naming_spaces(
    list_of_configuration_objects: list,
    nf_ea_com_universe: NfEaComUniverses,
):
    new_ea_objects_dictionary = (
        create_new_ea_objects_dictionary()
    )

    for (
        configuration_object
    ) in list_of_configuration_objects:
        __generalise_naming_space(
            configuration_object=configuration_object,
            nf_ea_com_universe=nf_ea_com_universe,
            new_ea_objects_dictionary=new_ea_objects_dictionary,
        )

    update_nf_ea_com_universe_with_dictionary(
        nf_ea_com_universe=nf_ea_com_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
    )


def __generalise_naming_space(
    configuration_object: GeneraliseNamesConfigurationObjects,
    nf_ea_com_universe: NfEaComUniverses,
    new_ea_objects_dictionary: dict,
):
    naming_space_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
        ea_guid=configuration_object.matched_naming_space.ea_guid,
    )

    named_object_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
        ea_guid=configuration_object.matched_named_object.ea_guid,
    )

    name_subtype_nf_uuid = get_intersection_of_dependency_and_association_linked(
        nf_ea_com_universe=nf_ea_com_universe,
        linked_by_dependency_nf_uuid=naming_space_nf_uuid,
        linked_by_association_nf_uuid=named_object_nf_uuid,
    )

    __add_generalisation(
        configuration_object=configuration_object,
        output_universe=nf_ea_com_universe,
        name_subtype_nf_uuid=name_subtype_nf_uuid,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
    )

    __add_association_name(
        configuration_object=configuration_object,
        output_universe=nf_ea_com_universe,
        named_object_nf_uuid=named_object_nf_uuid,
        name_subtype_nf_uuid=name_subtype_nf_uuid,
    )

    __add_stereotype(
        configuration_object=configuration_object,
        output_universe=nf_ea_com_universe,
        name_subtype_nf_uuid=name_subtype_nf_uuid,
    )


def __add_generalisation(
    configuration_object: GeneraliseNamesConfigurationObjects,
    output_universe: NfEaComUniverses,
    name_subtype_nf_uuid: str,
    new_ea_objects_dictionary: dict,
):
    name_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=output_universe,
        collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
        ea_guid=configuration_object.matched_name.ea_guid,
    )

    add_new_connector_to_dictionary(
        new_connector_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CONNECTORS
        ],
        place_1_nf_uuid=name_subtype_nf_uuid,
        place_2_nf_uuid=name_nf_uuid,
        connector_type=EaConnectorTypes.GENERALIZATION,
    )


def __add_association_name(
    configuration_object: GeneraliseNamesConfigurationObjects,
    output_universe: NfEaComUniverses,
    named_object_nf_uuid: str,
    name_subtype_nf_uuid: str,
):
    rename_connector(
        nf_ea_com_universe=output_universe,
        place_1_nf_uuid=name_subtype_nf_uuid,
        place_2_nf_uuid=named_object_nf_uuid,
        connector_type=EaConnectorTypes.ASSOCIATION,
        connector_name=configuration_object.matched_named_by_stereotype.object_name,
    )


def __add_stereotype(
    configuration_object: GeneraliseNamesConfigurationObjects,
    output_universe: NfEaComUniverses,
    name_subtype_nf_uuid: str,
):
    named_by_stereotype_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=output_universe,
        collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
        ea_guid=configuration_object.matched_named_by_stereotype.ea_guid,
    )

    add_stereotype_to_instances_associations(
        nf_ea_com_universe=output_universe,
        type_nf_uuid=name_subtype_nf_uuid,
        stereotype_nf_uuid=named_by_stereotype_nf_uuid,
    )



from abc import abstractmethod
from bclearer_orchestration_services.reporting_service.wrappers.run_and_log_function_wrapper import (
    run_and_log_function,
)


class BUnits:
    def __init__(
        self,
    ):
        pass

    def run(self) -> None:
        function_discriminator_name = (
            self.b_unit_domain_configuration.b_unit_bie_enum.b_enum_item_name
        )

        self.__run_b_unit(
            function_discriminator_name=function_discriminator_name
        )

    @run_and_log_function
    def __run_b_unit(self) -> None:
        self.b_unit_process_function()

    @abstractmethod
    def b_unit_process_function(
        self,
    ) -> ():
        pass



from bclearer_core.constants.standard_constants import (
    DEFAULT_FOREIGN_TABLE_SUFFIX,
    DEFAULT_MASTER_TABLE_SUFFIX,
    DEFAULT_NULL_VALUE,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_interop_services.dataframe_service.dataframe_helpers.dataframe_mergers import (
    inner_merge_dataframes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)


def get_mapped_nf_uuid_from_mapped_universe(
    source_nf_uuid: str,
    source_universe_collection_dictionary: dict,
    mapped_universe_collection_dictionary: dict,
) -> str:
    ea_guid_column_name = (
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
    )

    nf_uuids_column_name = (
        NfColumnTypes.NF_UUIDS.column_name
    )

    nf_uuids_source_universe_column_name = (
        "nf_uuids_source_universe"
    )

    nf_uuids_mapped_universe_column_name = (
        "nf_uuids_mapped_universe"
    )

    ea_guids_to_nf_uuids_source_universe = source_universe_collection_dictionary[
        NfEaComCollectionTypes.THIN_EA_EXPLICIT_OBJECTS
    ][
        [
            ea_guid_column_name,
            nf_uuids_column_name,
        ]
    ].copy()

    ea_guids_to_nf_uuids_source_universe.rename(
        columns={
            nf_uuids_column_name: nf_uuids_source_universe_column_name,
        },
        inplace=True,
    )

    ea_guids_to_nf_uuids_mapped_universe = mapped_universe_collection_dictionary[
        NfEaComCollectionTypes.THIN_EA_EXPLICIT_OBJECTS
    ][
        [
            ea_guid_column_name,
            nf_uuids_column_name,
        ]
    ].copy()

    ea_guids_to_nf_uuids_mapped_universe.rename(
        columns={
            nf_uuids_column_name: nf_uuids_mapped_universe_column_name,
        },
        inplace=True,
    )

    ea_guids_to_nf_uuids_map = inner_merge_dataframes(
        master_dataframe=ea_guids_to_nf_uuids_source_universe,
        master_dataframe_key_columns=[
            ea_guid_column_name,
        ],
        merge_suffixes=[
            DEFAULT_MASTER_TABLE_SUFFIX,
            DEFAULT_FOREIGN_TABLE_SUFFIX,
        ],
        foreign_key_dataframe=ea_guids_to_nf_uuids_mapped_universe,
        foreign_key_dataframe_fk_columns=[
            ea_guid_column_name,
        ],
        foreign_key_dataframe_other_column_rename_dictionary={
            nf_uuids_mapped_universe_column_name: nf_uuids_mapped_universe_column_name,
        },
    )

    mapped_nf_uuids_keyed_on_source_nf_uuids_dictionary = (
        {}
    )

    for (
        row_tuple
    ) in (
        ea_guids_to_nf_uuids_map.itertuples()
    ):
        row_source_nf_uuid = (
            row_tuple.nf_uuids_source_universe
        )

        row_mapped_nf_uuid = (
            row_tuple.nf_uuids_mapped_universe
        )

        mapped_nf_uuids_keyed_on_source_nf_uuids_dictionary[
            row_source_nf_uuid
        ] = row_mapped_nf_uuid

    if (
        source_nf_uuid
        in mapped_nf_uuids_keyed_on_source_nf_uuids_dictionary
    ):
        return mapped_nf_uuids_keyed_on_source_nf_uuids_dictionary[
            source_nf_uuid
        ]

    return DEFAULT_NULL_VALUE
